#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for annotate/annotation_functions.py
"""

import pytest
import os
import logging
import shutil

import test.test_unit.utilities_for_tests as tutil
import PanACoTA.utils as utils
from PanACoTA.annotate_module import annotation_functions as afunc


# Define variables used by several tests
DBDIR = os.path.join("test", "data", "annotate")
GEN_PATH = os.path.join(DBDIR, "genomes")
TMP_PATH = os.path.join(DBDIR, "tmp_files")
EXP_DIR = os.path.join(DBDIR, 'exp_files')
TEST_DIR = os.path.join(DBDIR, 'test_files')
GENEPATH = os.path.join(DBDIR, "generated_by_unit-tests")
LOGFILE_BASE = os.path.join(GENEPATH, "logfile.log")
LOGFILES = [LOGFILE_BASE + ext for ext in [".log", ".log.debug", ".log.details", ".log.err"]]


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    if os.path.isdir(GENEPATH):
        content = os.listdir(GENEPATH)
        for f in content:
            assert f.startswith(".fuse")
    else:
        os.mkdir(GENEPATH)
    print("setup")

    yield
    for f in LOGFILES:
        if os.path.exists(f):
            os.remove(f)
    shutil.rmtree(GENEPATH, ignore_errors=True)
    print("teardown")


# Create a logger with sublogger inside
def my_logger(name):
    """
    logger given to function called by a subprocess
    """
    import multiprocessing
    m = multiprocessing.Manager()
    q = m.Queue()
    qh = logging.handlers.QueueHandler(q)
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    root.handlers = []
    logging.addLevelName(utils.detail_lvl(), "DETAIL")
    root.addHandler(qh)
    return q, logging.getLogger(name)


def test_prodigal_train(caplog):
    """
    Check prodigal training on a genome
    """
    caplog.set_level(logging.DEBUG)
    train_gpath = os.path.join(GEN_PATH, "A_H738.fasta")
    gtrain = afunc.prodigal_train(train_gpath, GENEPATH)
    assert ("prodigal command: prodigal -i test/data/annotate/genomes/A_H738.fasta "
            "-t test/data/annotate/generated_by_unit-tests/A_H738.fasta.trn") in caplog.text
    assert ("End annotating A_H738.fasta (from test/data/annotate/genomes/A_H738.fasta)")
    assert gtrain == ("test/data/annotate/generated_by_unit-tests/A_H738.fasta.trn")


def test_prodigal_train_error(caplog):
    """
    Check prodigal training on a genome too small
    """
    caplog.set_level(logging.DEBUG)
    train_gpath = os.path.join(GEN_PATH, "H299_H561.fasta")
    gtrain = afunc.prodigal_train(train_gpath, GENEPATH)
    assert ("prodigal command: prodigal -i test/data/annotate/genomes/H299_H561.fasta "
            "-t test/data/annotate/generated_by_unit-tests/H299_H561.fasta.trn") in caplog.text
    assert ("Error while trying to train prodigal on H299_H561.fasta") in caplog.text
    assert gtrain == ""


def test_prodigal_train_exists(caplog):
    """
    Check prodigal training when the training file already exists
    """
    caplog.set_level(logging.DEBUG)
    train_gpath = os.path.join(GEN_PATH, "H299_H561.fasta")
    # Create (empty) trn file
    gtrain = os.path.join(GENEPATH, "H299_H561.fasta.trn")
    open(gtrain, "w").close()
    gtrain_found = afunc.prodigal_train(train_gpath, GENEPATH)
    assert gtrain_found == gtrain
    assert ("A training file already exists (test/data/annotate/generated_by_unit-tests/"
            "H299_H561.fasta.trn). It will be used to annotate all genomes.") in caplog.text


def test_check_prodigal_nofaa():
    """
    Check that check_prodigal returns false when a faa file is missing, and an error message
    """
    logger = my_logger("test_check_prodigal_nofaa")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prokkaRes")
    ori_name = "prokka_out_for_test"
    out_dir = os.path.join(GENEPATH, "out_test_nofaa")
    name = "prodigal_out_for_test-missfaa"
    os.makedirs(out_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".tbl"),
                    os.path.join(out_dir, name + ".tbl"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))
    logf = os.path.join(GENEPATH, "prodigal.log")
    gpath = "path/to/nogenome/original_name.fna"
    nbcont = 7
    assert not afunc.check_prodigal(gpath, name, out_dir, logger[1])
    msg = "prodigal_out_for_test-missfaa original_name.fna: no or several .faa file(s)"
    q = logger[0]
    assert q.qsize() == 1
    assert q.get().message == msg


def test_check_prodigal_sevfaa():
    """
    Check that check_prodigal returns false when there is more than 1 faa file,
    and an error message
    """
    logger = my_logger("test_check_prodigal_sevfaa")
    ori_prod_dir = os.path.join(TEST_DIR, "original_name.fna-prodigalRes")
    ori_name = "prodigal.outtest.ok"
    out_dir = os.path.join(GENEPATH, "out_test_nofaa")
    name = "prodigal_out_for_test-missfaa"
    os.makedirs(out_dir)
    shutil.copyfile(os.path.join(ori_prod_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    shutil.copyfile(os.path.join(ori_prod_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prod_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + "2.faa"))
    shutil.copyfile(os.path.join(ori_prod_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))
    logf = os.path.join(GENEPATH, "prodigal.log")
    gpath = "path/to/nogenome/original_name.fna"
    nbcont = 7
    assert not afunc.check_prodigal(gpath, name, out_dir, logger[1])
    msg = "prodigal_out_for_test-missfaa original_name.fna: no or several .faa file(s)"
    q = logger[0]
    assert q.qsize() == 1
    assert q.get().message == msg


def test_check_prodigal_noffn():
    """
    Check that check_prodigal returns false when a ffn file is missing, and an error message
    """
    logger = my_logger("test_check_prodigal_noffn")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prodigalRes")
    ori_name = "prodigal.outtest.ok"
    out_dir = os.path.join(GENEPATH, "out_test_noffn")
    name = "prodigal_out_for_test-missffn"
    os.makedirs(out_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))
    logf = os.path.join(GENEPATH, "prodigal.log")
    gpath = "path/to/nogenome/original_name.fna"
    nbcont = 7
    assert not afunc.check_prodigal(gpath, name, out_dir, logger[1])
    msg = "prodigal_out_for_test-missffn original_name.fna: no or several .ffn file"
    q = logger[0]
    assert q.qsize() == 1
    assert msg in q.get().message


def test_check_prodigal_sevffn():
    """
    Check that check_prodigal returns false when there is more than 1 ffn file,
    and an error message
    """
    logger = my_logger("test_check_prodigal_sevffn")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prodigalRes")
    ori_name = "prodigal.outtest.ok"
    out_dir = os.path.join(GENEPATH, "out_test_noffn")
    name = "prodigal_out_for_test-sevffn"
    os.makedirs(out_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + "2.ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))
    logf = os.path.join(GENEPATH, "prodigal.log")
    gpath = "path/to/nogenome/original_name.fna"
    nbcont = 7
    assert not afunc.check_prodigal(gpath, name, out_dir, logger[1])
    msg = "prodigal_out_for_test-sevffn original_name.fna: no or several .ffn file"
    q = logger[0]
    assert q.qsize() == 1
    assert msg in q.get().message


def test_check_prodigal_nogff():
    """
    Check that check_prodigal returns false when a ffn file is missing, and an error message
    """
    logger = my_logger("test_check_prodigal_nogff")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prodigalRes")
    ori_name = "prodigal.outtest.ok"
    out_dir = os.path.join(GENEPATH, "out_test_noffn")
    name = "prodigal_out_for_test-missgff"
    os.makedirs(out_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    logf = os.path.join(GENEPATH, "prodigal.log")
    gpath = "path/to/nogenome/original_name.fna"
    nbcont = 7
    assert not afunc.check_prodigal(gpath, name, out_dir, logger[1])
    msg = "prodigal_out_for_test-missgff original_name.fna: no or several .gff file"
    q = logger[0]
    assert q.qsize() == 1
    assert msg in q.get().message


def test_check_prodigal_sevgff():
    """
    Check that check_prodigal returns false when there is more than 1 ffn file,
    and an error message
    """
    logger = my_logger("test_check_prodigal_sevgff")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prodigalRes")
    ori_name = "prodigal.outtest.ok"
    out_dir = os.path.join(GENEPATH, "out_test_noffn")
    name = "prodigal_out_for_test-sevgff"
    os.makedirs(out_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + "2.gff"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))
    logf = os.path.join(GENEPATH, "prodigal.log")
    gpath = "path/to/nogenome/original_name.fna"
    assert not afunc.check_prodigal(gpath, name, out_dir, logger[1])
    msg = "prodigal_out_for_test-sevgff original_name.fna: no or several .gff file"
    q = logger[0]
    assert q.qsize() == 1
    assert msg in q.get().message


def test_check_prodigal_emptyfaa():
    """
    Check that check_prodigal returns false when there are all expected files, but faa
    file is empty
    """
    logger = my_logger("test_check_prodigal_ok")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prodigalRes")
    ori_name = "prodigal.outtest.ok"
    out_dir = os.path.join(GENEPATH, "out_test_noffn")
    name = "prodigal_out_for_test"
    logf = os.path.join(GENEPATH, "prodigal.log")
    os.makedirs(out_dir)
    open(os.path.join(out_dir, name + ".faa"), "w").close()
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))
    gpath = "path/to/nogenome/original_name.fna"
    assert not afunc.check_prodigal(gpath, name, out_dir, logger[1])
    msg = "Genome prodigal_out_for_test (from original_name.fna): At least one of your Prodigal result file is empty."
    q = logger[0]
    assert q.qsize() == 1
    assert msg in q.get().message


def test_check_prodigal_emptyffn():
    """
    Check that check_prodigal returns false when there are all expected files, but ffn
    file is empty
    """
    logger = my_logger("test_check_prodigal_ok")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prodigalRes")
    ori_name = "prodigal.outtest.ok"
    out_dir = os.path.join(GENEPATH, "out_test_noffn")
    name = "prodigal_out_for_test"
    logf = os.path.join(GENEPATH, "prodigal.log")
    os.makedirs(out_dir)
    open(os.path.join(out_dir, name + ".ffn"), "w").close()
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))
    gpath = "path/to/nogenome/original_name.fna"
    assert not afunc.check_prodigal(gpath, name, out_dir, logger[1])
    msg = "Genome prodigal_out_for_test (from original_name.fna): At least one of your Prodigal result file is empty."
    q = logger[0]
    assert q.qsize() == 1
    assert msg in q.get().message


def test_check_prodigal_emptygff():
    """
    Check that check_prodigal returns false when there are all expected files, but gff
    file is empty
    """
    logger = my_logger("test_check_prodigal_ok")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prodigalRes")
    ori_name = "prodigal.outtest.ok"
    out_dir = os.path.join(GENEPATH, "out_test_noffn")
    name = "prodigal_out_for_test"
    logf = os.path.join(GENEPATH, "prodigal.log")
    os.makedirs(out_dir)
    open(os.path.join(out_dir, name + ".gff"), "w").close()
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    gpath = "path/to/nogenome/original_name.fna"
    assert not afunc.check_prodigal(gpath, name, out_dir, logger[1])
    msg = "Genome prodigal_out_for_test (from original_name.fna): At least one of your Prodigal result file is empty."
    q = logger[0]
    assert q.qsize() == 1
    assert msg in q.get().message


def test_check_prodigal_ok():
    """
    Check that everything is ok with prodigal results (tbl, faa and ffn files exist,
    and number of CDS, CRISPR and genes correspond between them)
    """
    logger = my_logger("test_check_prodigal_ok")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prodigalRes")
    ori_name = "prodigal.outtest.ok"
    out_dir = os.path.join(GENEPATH, "out_test_noffn")
    name = "prodigal_out_for_test"
    logf = os.path.join(GENEPATH, "prodigal.log")
    os.makedirs(out_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))
    gpath = "path/to/nogenome/original_name.fna"
    assert afunc.check_prodigal(gpath, name, out_dir, logger[1])


def test_run_prodigal_out_exists_ok():
    """
    Test that when the output directory already exists, and files inside are OK,
    run_prodigal returns True, with a warning message indicating that prodigal did not rerun.
    """
    logger = my_logger("test_run_prodigal_out_exists_ok")
    utils.init_logger(LOGFILE_BASE, 0, 'prodigal_out_exists_ok')
    gpath = "path/to/nogenome/original_name.fna"
    cores_prodigal = 1
    name = "prodigal.outtest.ok"
    force = False
    nbcont = 7
    trn_file = os.path.join(TEST_DIR, "A_H738-and-B2_A3_5.fna.trn")
    arguments = (gpath, TEST_DIR, cores_prodigal, name, force, nbcont, trn_file, logger[0])
    assert afunc.run_prodigal(arguments)

    q = logger[0]
    assert q.qsize() == 4
    # start annotating :
    assert q.get().message.startswith("Start annotating prodigal.outtest.ok (from "
                                      "path/to/nogenome/original_name.fna sequence) with Prodigal")
    # # warning prodigal results folder exists:
    assert q.get().message.startswith("Prodigal results folder test/data/annotate/test_files/"
                                      "original_name.fna-prodigalRes already exists.")
    # Results in result folder are ok
    assert q.get().message.startswith("Prodigal did not run again. Formatting step will use "
                                      "already generated results of Prodigal in "
                                      "test/data/annotate/test_files/"
                                      "original_name.fna-prodigalRes.")
    # End annotation:
    assert q.get().message.startswith("End annotating")


def test_run_prodigal_out_exists_error():
    """
    Test that when the output directory already exists, and 1 file is missing,
    run_prodigal returns False, and writes the warning message saying that prodigal did not
    rerun, + the warning message for the missing file(s).
    """
    logger = my_logger("test_run_prodigal_out_exists_error")
    utils.init_logger(LOGFILE_BASE, 0, 'prodigal_out_error')
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prodigalRes")
    ori_name = "prodigal.outtest.ok"
    new_prok_dir = os.path.join(GENEPATH, "original_name-error-prodigalRes")
    name = "prodigal_out_for_test-wrongCDS"
    os.makedirs(new_prok_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(new_prok_dir, name + ".ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(new_prok_dir, name + ".faa"))
    open(os.path.join(new_prok_dir, name + ".gff"), "w").close()
    gpath = "path/to/nogenome/original_name-error"
    cores_prodigal = 1
    force = False
    trn_file = os.path.join(TEST_DIR, "A_H738-and-B2_A3_5.fna.trn")
    nbcont = 7
    arguments = (gpath, GENEPATH, cores_prodigal, name, force, nbcont, trn_file, logger[0])
    assert not afunc.run_prodigal(arguments)
    q = logger[0]
    assert q.qsize() == 4
    # start annotating :
    assert q.get().message.startswith("Start annotating")
    # warning prodigal results folder exists:
    assert q.get().message == ("Prodigal results folder test/data/annotate/"
                               "generated_by_unit-tests/"
                               "original_name-error-prodigalRes already exists.")
    # error, empty gff
    msg = ("Genome prodigal_out_for_test-wrongCDS (from original_name-error): "
           "At least one of your Prodigal result file is empty.")
    assert q.get().message == msg
    # warning, files in outdir are not as expected
    assert q.get().message.startswith("Problems in the files contained in your already existing "
                                      "output dir (test/data/annotate/generated_by_unit-tests/"
                                      "original_name-error-prodigalRes")


def test_run_prodigal_out_exists_force():
    """
    Test that when the output directory already exists with wrong files, but force is on,
    prodigal is rerun and outputs the right files
    """
    logger = my_logger("test_run_prodigal_out_exists_force")
    utils.init_logger(LOGFILE_BASE, 0, 'force')
    gpath = os.path.join(GEN_PATH, "H299_H561.fasta")
    out_prokdir = os.path.join(GENEPATH, "H299_H561.fasta-prodigalRes")
    name = "test_runprodigal_H299"
    # Put empty tbl, faa, ffn files in prodigal output dir, to check that they are overridden
    os.makedirs(out_prokdir)
    open(os.path.join(out_prokdir, name + ".gff"), "w").close()
    open(os.path.join(out_prokdir, name + ".faa"), "w").close()
    open(os.path.join(out_prokdir, name + ".ffn"), "w").close()
    cores_prodigal = 2
    force = True
    nbcont = 3
    trn_file = os.path.join(TEST_DIR, "A_H738-and-B2_A3_5.fna.trn")
    arguments = (gpath, GENEPATH, cores_prodigal, name, force, nbcont, trn_file, logger[0])
    assert afunc.run_prodigal(arguments)
    # As we used 'force', tbl, faa and ffn files, which were empty, must have been replaced
    # by the prodigal output
    exp_dir = os.path.join(EXP_DIR, "H299_H561.fasta-prodigalRes",
                           "ESCO.1015.00001")
    out_gff = os.path.join(out_prokdir, name + ".gff")
    out_faa = os.path.join(out_prokdir, name + ".faa")
    out_ffn = os.path.join(out_prokdir, name + ".ffn")
    # Check that faa and ffn files are as expected
    assert os.path.isfile(out_faa)
    assert tutil.compare_order_content(exp_dir + ".faa", out_faa)
    assert os.path.isfile(out_ffn)
    assert tutil.compare_order_content(exp_dir + ".ffn", out_ffn)
    q = logger[0]
    assert q.qsize() == 4
    assert q.get() .message.startswith("Prodigal results folder already exists, but is "
                                       "removed because --force option was used")
    assert q.get() .message.startswith("Start annotating test_runprodigal_H299 (from test/data/"
                                       "annotate/genomes/H299_H561.fasta sequence) "
                                       "with Prodigal")
    assert q.get().message.startswith("Prodigal command: prodigal -i test/data/annotate/genomes/"
                                      "H299_H561.fasta -d test/data/annotate/"
                                      "generated_by_unit-tests/H299_H561.fasta-prodigalRes/"
                                      "test_runprodigal_H299.ffn -a test/data/annotate/"
                                      "generated_by_unit-tests/H299_H561.fasta-prodigalRes/"
                                      "test_runprodigal_H299.faa -f gff -o test/data/annotate/"
                                      "generated_by_unit-tests/H299_H561.fasta-prodigalRes/"
                                      "test_runprodigal_H299.gff -t "
                                      "test/data/annotate/test_files/A_H738-and-B2_A3_5.fna.trn "
                                      "-q")
    assert q.get() .message.startswith("End annotating test_runprodigal_H299 "
                                       "(from test/data/annotate/genomes/H299_H561.fasta)")


def test_run_prodigal_out_doesnt_exist():
    """
    Test that when the output directory does not exist, it creates it, and runs prodigal
    with all expected outfiles
    """
    logger = my_logger("test_run_prodigal_out_doesnt_exist")
    utils.init_logger(LOGFILE_BASE, 0, 'test_run_prodigal_out_doesnt_exist')
    gpath = os.path.join(GEN_PATH, "H299_H561.fasta")
    out_dir = os.path.join(GENEPATH, "H299_H561.fasta-prodigalRes")
    cores_prodigal = 2
    name = "test_runprodigal_H299"
    force = False
    trn_file = os.path.join(TEST_DIR, "A_H738-and-B2_A3_5.fna.trn")
    nbcont = 3
    arguments = (gpath, GENEPATH, cores_prodigal, name, force, nbcont, trn_file, logger[0])
    assert afunc.run_prodigal(arguments)
    # Check content of tbl, ffn and faa files
    exp_dir = os.path.join(EXP_DIR, "H299_H561.fasta-prodigalRes",
                           "ESCO.1015.00001")
    out_faa = os.path.join(out_dir, name + ".faa")
    out_ffn = os.path.join(out_dir, name + ".ffn")
    out_gff = os.path.join(out_dir, name + ".gff")
    # Check that faa and ffn files are as expected
    assert os.path.isfile(out_faa)
    assert tutil.compare_order_content(exp_dir + ".faa", out_faa)
    assert os.path.isfile(out_ffn)
    assert tutil.compare_order_content(exp_dir + ".ffn", out_ffn)
    assert os.path.isfile(out_ffn)
    assert tutil.compare_order_content(exp_dir + ".gff", out_gff)
    q = logger[0]
    assert q.qsize() == 3
    assert q.get().message.startswith("Start annotating")
    assert q.get().message == ("Prodigal command: prodigal -i test/data/annotate/genomes/"
                               "H299_H561.fasta -d test/data/annotate/"
                               "generated_by_unit-tests/H299_H561.fasta-prodigalRes/"
                               "test_runprodigal_H299.ffn -a test/data/annotate/"
                               "generated_by_unit-tests/H299_H561.fasta-prodigalRes/"
                               "test_runprodigal_H299.faa -f gff -o test/data/annotate/"
                               "generated_by_unit-tests/H299_H561.fasta-prodigalRes/"
                               "test_runprodigal_H299.gff -t "
                               "test/data/annotate/test_files/A_H738-and-B2_A3_5.fna.trn "
                               "-q")
    assert q.get().message.startswith("End annotating")


def test_run_prodigal_small():
    """
    Test that when the output directory does not exist, it creates it, and runs prodigal
    with all expected outfiles. Here, we run prodigal with --small option (on a small genome)
    """
    logger = my_logger("test_run_prodigal_small")
    utils.init_logger(LOGFILE_BASE, 0, 'test_run_prodigal_small')
    gpath = os.path.join(GEN_PATH, "H299_H561.fasta")
    out_dir = os.path.join(GENEPATH, "H299_H561.fasta-prodigalRes")
    cores_prodigal = 2
    name = "test_runprodigal_small_H299"
    force = False
    trn_file = "small option"
    nbcont = 3
    arguments = (gpath, GENEPATH, cores_prodigal, name, force, nbcont, trn_file, logger[0])
    assert afunc.run_prodigal(arguments)

    # Check content of tbl, ffn and faa files
    exp_dir = os.path.join(EXP_DIR, "H299_H561.fasta_small-prodigalRes",
                           "test_runprodigal_small_H299")
    out_faa = os.path.join(out_dir, name + ".faa")
    out_ffn = os.path.join(out_dir, name + ".ffn")
    out_gff = os.path.join(out_dir, name + ".gff")
    # Check that faa and ffn files are as expected
    assert os.path.isfile(out_faa)
    assert tutil.compare_order_content(exp_dir + ".faa", out_faa)
    assert os.path.isfile(out_ffn)
    assert tutil.compare_order_content(exp_dir + ".ffn", out_ffn)
    assert os.path.isfile(out_ffn)
    assert tutil.compare_order_content(exp_dir + ".gff", out_gff)
    # Check logs
    q = logger[0]
    assert q.qsize() == 3
    assert q.get().message.startswith("Start annotating")
    prodigal_cmd = q.get().message
    assert ("Prodigal command: prodigal -i test/data/annotate/genomes/"
            "H299_H561.fasta -d test/data/annotate/"
            "generated_by_unit-tests/H299_H561.fasta-prodigalRes/"
            "test_runprodigal_small_H299.ffn -a test/data/annotate/"
            "generated_by_unit-tests/H299_H561.fasta-prodigalRes/"
            "test_runprodigal_small_H299.faa -f gff -o test/data/annotate/"
            "generated_by_unit-tests/H299_H561.fasta-prodigalRes/"
            "test_runprodigal_small_H299.gff -p meta -q")in prodigal_cmd
    assert q.get().message.startswith("End annotating")


def test_run_prodigal_out_problem_running():
    """
    Check that when a problem occurs while trying to run prodigal, run_prodigal returns False,
    and the error message indicating to read in the log why it couldn't run
    """
    logger = my_logger("test_run_prodigal_out_problem_running")
    utils.init_logger(LOGFILE_BASE, 0, 'test_run_prodigal_out_problem_running')
    gpath = os.path.join(GEN_PATH, "H299_H561bis.fasta")
    cores_prodigal = 2
    name = "test_runprodigal_H299-error"
    force = False
    nbcont = 3
    trn_file = os.path.join(TEST_DIR, "A_H738-and-B2_A3_5.fna.trn")
    logf = os.path.join(GENEPATH, "H299_H561bis.fasta-prodigal.log")
    arguments = (gpath, GENEPATH, cores_prodigal, name, force, nbcont, trn_file, logger[0])
    assert not afunc.run_prodigal(arguments)
    # Check that output directory is empty
    outdir = os.path.join(GENEPATH, "H299_H561bis.fasta-prodigalRes")
    assert os.listdir(outdir) == []
    # Check logs
    q = logger[0]
    assert q.qsize() == 3
    assert q.get().message.startswith("Start annotating")
    assert q.get().message.startswith("Prodigal command: prodigal -i test/data/annotate/genomes/"
                               "H299_H561bis.fasta -d test/data/annotate/"
                               "generated_by_unit-tests/H299_H561bis.fasta-prodigalRes/"
                               "test_runprodigal_H299-error.ffn -a test/data/annotate/"
                               "generated_by_unit-tests/H299_H561bis.fasta-prodigalRes/"
                               "test_runprodigal_H299-error.faa -f gff -o test/data/annotate/"
                               "generated_by_unit-tests/H299_H561bis.fasta-prodigalRes/"
                               "test_runprodigal_H299-error.gff -t "
                               "test/data/annotate/test_files/A_H738-and-B2_A3_5.fna.trn "
                               "-q")
    assert q.get().message.startswith("Error while trying to run prodigal. See test/data/"
                                      "annotate/generated_by_unit-tests/"
                                      "H299_H561bis.fasta-prodigal.log.err.")


def test_run_prodigal_noout_notrain():
    """
    Prodigal result directory does not exist (not already run)
    training file does not exist (probably, problem while trying to train)
    -> return  False
    """
    logger = my_logger("test_run_prodigal_out_exists_error")
    utils.init_logger(LOGFILE_BASE, 0, 'prodigal_out_error')
    gpath = "path/to/nogenome/original_name-error"
    cores_prodigal = 1
    name = "prodigal_out_for_test-wrongCDS"
    force = False
    nbcont = 7
    trn_file = "ghost_trn_file"
    arguments = (gpath, GENEPATH, cores_prodigal, name, force, nbcont, trn_file, logger[0])
    assert not afunc.run_prodigal(arguments)
    q = logger[0]
    assert q.qsize() == 0



#!/usr/bin/env python3
# coding: utf-8

"""
Functional tests for PanACoTA 'prepare' subcommand
"""

from PanACoTA.subcommands import prepare
import test.test_unit.utilities_for_tests as tutil

import pytest
import os
import subprocess
import shutil
import time
import argparse
import logging
import glob


# LOGFILE_BASE = "test_main_from_parse"
# Define variables used by several tests
DBDIR = os.path.join("test", "data", "prepare")
GEN_PATH = os.path.join(DBDIR, "genomes")
TEST_DIR = os.path.join(DBDIR, 'test_files')
GENEPATH = os.path.join(DBDIR, "generated_by_func-tests")


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    if not os.path.isdir(GENEPATH):
        print("setup")
        os.mkdir(GENEPATH)
    print("setup")

    yield
    shutil.rmtree(GENEPATH, ignore_errors=True)
    print("teardown")


def test_main_from_parse():
    """
    Run
    """
    args = argparse.Namespace()
    args.argv = ["prepare", "test_func_prepare"]
    args.ncbi_species_name = "Acetobacter orleanensis"
    args.ncbi_species_taxid = "104099"
    args.ncbi_taxid = ""
    args.ncbi_section = "refseq"
    args.outdir = GENEPATH
    args.tmp_dir = ""
    args.parallel = 1
    args.norefseq = False
    args.db_dir = ""
    args.only_mash = False
    args.info_file = ""
    args.l90 = 100
    args.nbcont = 999
    args.cutn = 0
    args.min_dist = 1e-4
    args.max_dist = 0.06
    args.verbose = 0
    args.quiet = False
    args.levels = ""

    prepare.main_from_parse(args)

    # Check output files
    summary =  os.path.join(GENEPATH, "assembly_summary-Acetobacter_orleanensis.txt")
    assert os.path.isfile(summary)
    # Check that the NCBI_genome_download output directory exists
    ngd_outdir = os.path.join(GENEPATH, "refseq", "bacteria")
    # And that it contains folders
    assert os.path.isdir(ngd_outdir)
    assert len(os.listdir(ngd_outdir)) >= 4
    # Check logfiles are here
    log_files = glob.glob(os.path.join(GENEPATH, "*log*"))
    assert len(log_files) == 3
    # Check tmp files folder created, but empty as we do not split
    tmp_folder = os.listdir(os.path.join(GENEPATH, "tmp_files"))
    assert len(tmp_folder) == 0
    # Check Database_init folder created, with at list 4 ".fna" genomes
    fna_files = glob.glob(os.path.join(GENEPATH, "Database_init", "*.fna"))
    assert len(fna_files) >= 4


def test_main_from_parse_longspeciesname():
    """
    Run
    """
    args = argparse.Namespace()
    args.argv = ["prepare", "test_func_prepare"]
    args.ncbi_species_name = "Salmonella enterica subsp. enterica serovar Paratyphi C"
    args.ncbi_species_taxid = ""
    args.ncbi_taxid = ""
    args.ncbi_section = "refseq"
    args.outdir = GENEPATH
    args.tmp_dir = ""
    args.parallel = 1
    args.norefseq = False
    args.db_dir = ""
    args.only_mash = False
    args.info_file = ""
    args.l90 = 100
    args.nbcont = 999
    args.cutn = 0
    args.min_dist = 1e-4
    args.max_dist = 0.06
    args.verbose = 0
    args.quiet = False
    args.levels = ""

    prepare.main_from_parse(args)

    # Check output files
    summary =  os.path.join(GENEPATH, "assembly_summary-Salmonella_enterica_subsp._enterica_serovar_Paratyphi_C.txt")
    assert os.path.isfile(summary)
    # Check that the NCBI_genome_download output directory exists
    ngd_outdir = os.path.join(GENEPATH, "refseq", "bacteria")
    # And that it contains folders
    assert os.path.isdir(ngd_outdir)
    assert len(os.listdir(ngd_outdir)) >= 1
    # Check logfiles are here
    log_files = glob.glob(os.path.join(GENEPATH, "*log*"))
    assert len(log_files) == 3
    # Check tmp files folder created, but empty as we do not split
    tmp_folder = os.listdir(os.path.join(GENEPATH, "tmp_files"))
    assert len(tmp_folder) == 0
    # Check Database_init folder created, with at list 4 ".fna" genomes
    fna_files = glob.glob(os.path.join(GENEPATH, "Database_init", "*.fna"))
    assert len(fna_files) >= 1


def test_main_not_only_mash_infoexists():
    """
    We run without option only_mash, but still provide a lstinfo file
    -> will change its name to .back to save it when the new file will be created
    """
    NCBI_species_name = ""
    NCBI_species_taxid = "104099"
    NCBI_taxid = ""
    NCBI_section = "refseq"
    levels = ""
    outdir = GENEPATH
    tmp_dir = os.path.join(outdir, "temporary_directory")
    threads = 1
    norefseq = False
    db_dir = ""
    only_mash = False
    info_file = os.path.join(outdir, "LSTINFO-existing.lst")
    open(info_file, "w").close()  #create empty info file, to check it is renamed
    l90 = 100
    nbcont = 999
    cutn = 5
    min_dist = 1e-4
    max_dist = 0.06
    verbose = 2
    quiet = False
    out_info_file = os.path.join(outdir, "LSTINFO-104099-filtered-0.0001_0.06.txt")
    assert prepare.main("cmd", NCBI_species_name, NCBI_species_taxid, NCBI_taxid, levels, NCBI_section, outdir, tmp_dir,
                        threads, norefseq, db_dir, only_mash, info_file, l90, nbcont,
                        cutn, min_dist, max_dist, verbose, quiet) == out_info_file

    # Check output files
    summary =  os.path.join(GENEPATH, "assembly_summary-104099.txt")
    assert os.path.isfile(summary)
    # Check that the NCBI_genome_download output directory exists
    ngd_outdir = os.path.join(GENEPATH, "refseq", "bacteria")
    # And that it contains folders
    assert os.path.isdir(ngd_outdir)
    assert len(os.listdir(ngd_outdir)) >= 4
    # Check logfiles are here
    log_files = glob.glob(os.path.join(GENEPATH, "*log*"))
    assert len(log_files) == 3
    # Check tmp files folder created, but empty as we do not split
    tmp_files = glob.glob(os.path.join(tmp_dir, "*.fna_prepare-split5N.fna"))
    assert len(tmp_files) >= 4
    # Check Database_init folder created, with at list 4 ".fna" genomes
    fna_files = glob.glob(os.path.join(GENEPATH, "Database_init", "*.fna"))
    assert len(fna_files) >= 4
    # Check that LSTINFO file existing was renamed and still empty
    # And new LSTINFO file created
    assert os.path.isfile(info_file + ".back")
    assert os.stat(info_file + ".back").st_size == 0


def test_main_wrong_taxid(capsys):
    """
    We run without option only_mash, but still provide a lstinfo file
    -> will change its name to .back to save it when the new file will be created
    """
    NCBI_species_name = ""
    NCBI_taxid = "123"
    NCBI_species_taxid = ""
    NCBI_section = "genbank"
    levels = ""
    outdir = ""
    tmp_dir = os.path.join("123", "temporary_directory")
    threads = 1
    norefseq = False
    info_file = ""
    db_dir = ""
    only_mash = False
    l90 = 100
    nbcont = 999
    cutn = 5
    min_dist = 1e-4
    max_dist = 0.06
    verbose = 2
    quiet = False
    res_outdir = "123"
    with pytest.raises(SystemExit):
        prepare.main("cmd", NCBI_species_name, NCBI_species_taxid, NCBI_taxid, levels, NCBI_section, 
                     outdir, tmp_dir, threads, norefseq,
                     db_dir, only_mash, info_file, l90, nbcont, cutn, min_dist, max_dist,
                     verbose, quiet)
    _, err = capsys.readouterr()
    assert ("Could not download genomes. Check that you gave valid NCBI taxid and/or "
            "NCBI species name. If you gave both, check that given taxID and name really "
            "correspond to the same species.") in err
    # Check output files
    summary =  os.path.join(res_outdir, "assembly_summary-123.txt")
    assert not os.path.isfile(summary)
    ngd_outdir = os.path.join(res_outdir, "genbank", "bacteria")
    assert not os.path.isdir(ngd_outdir)
    # # Check logfiles are here
    log_files = glob.glob(os.path.join(res_outdir, "*log*"))
    assert len(log_files) == 3
    # Check tmp files folder created, but empty asnothing is downloaded
    assert len(os.listdir(tmp_dir)) == 0
    # Check Database_init folder created, with at list 4 ".fna" genomes
    assert not os.path.isdir(os.path.join(res_outdir, "Database_init"))

    # Remove output directory
    shutil.rmtree(res_outdir, ignore_errors=True)


def test_main_norefseq_wrongdbpath(capsys):
    """
    We run with option norefseq, but given db_dir does not exist.
    -> error message
    """
    NCBI_species_name = ""
    NCBI_species_taxid = ""
    NCBI_taxid = ""
    NCBI_section = "refseq"
    levels = ""
    outdir = GENEPATH
    tmp_dir = os.path.join(outdir, "temporary_directory")
    threads = 1
    norefseq = True
    db_dir = "dbdir"
    only_mash = False
    l90 = 100
    nbcont = 999
    cutn = 5
    min_dist = 1e-4
    max_dist = 0.06
    verbose = 15
    quiet = False
    info_file = ""
    with pytest.raises(SystemExit):
        prepare.main("cmd", NCBI_species_name, NCBI_species_taxid, NCBI_taxid, levels, NCBI_section,
                     outdir, tmp_dir, threads, norefseq,
                     db_dir, only_mash, info_file, l90, nbcont, cutn, min_dist, max_dist,
                     verbose, quiet)
    _, err = capsys.readouterr()
    assert ("You asked to skip refseq downloads") in err
    assert ("Database folder dbdir supposed to contain fasta sequences does not exist. Please "
            "give a valid folder, or leave the default directory (no '-d' option)") in err
    # Check output files
    summary =  os.path.join(GENEPATH, "assembly_summary-123.txt")
    assert not os.path.isfile(summary)
    ngd_outdir = os.path.join(GENEPATH, "refseq", "bacteria")
    assert not os.path.isdir(ngd_outdir)
    # Check logfiles are here
    log_files = glob.glob(os.path.join(GENEPATH, "*log*"))
    assert len(log_files) == 4  #.log.debug as we put verbose = 15
    # Check tmp files folder created, but empty asnothing is downloaded
    assert len(os.listdir(tmp_dir)) == 0
    # Check Database_init folder created, with at list 4 ".fna" genomes
    assert not os.path.isdir(os.path.join(GENEPATH, "Database_init"))


def test_main_norefseq_nodefault_dbdir_nor_refseq(capsys):
    """
    We run with option norefseq, but given db_dir does not exist.
    -> error message
    """
    NCBI_species_name = ""
    NCBI_species_taxid = ""
    NCBI_taxid = ""
    NCBI_section = "genbank"
    levels = ""
    outdir = GENEPATH
    tmp_dir = ""
    threads = 1
    norefseq = True
    db_dir = ""
    only_mash = False
    l90 = 100
    nbcont = 999
    cutn = 5
    min_dist = 1e-4
    max_dist = 0.06
    verbose = 2
    quiet = False
    info_file = ""
    with pytest.raises(SystemExit):
        prepare.main("cmd", NCBI_species_name, NCBI_species_taxid, NCBI_taxid, levels, 
                     NCBI_section, outdir, tmp_dir, threads, norefseq,
                     db_dir, only_mash, info_file, l90, nbcont, cutn, min_dist, max_dist,
                     verbose, quiet)
    _, err = capsys.readouterr()
    assert ("You asked to skip genbank downloads") in err
    assert ("Database folder test/data/prepare/generated_by_func-tests/Database_init supposed "
            "to contain fasta sequences does not exist. We will check if the download folder "
            "(with compressed sequences) exists.") in err
    assert ("Folder test/data/prepare/generated_by_func-tests/genbank/bacteria "
            "does not exist. You do not have any genome to analyse. Possible reasons:\n") in err
    assert ("- if you want to rerun analysis in the same folder as "
            "sequences were downloaded (my_outdir/Database_init or "
            "my_outdir/genbank), make sure you have '-o my_outdir' option\n") in err
    assert ("- if you want to rerun analysis and save them in a new "
            "output folder called 'new_outdir', make sure you have '-o new_outdir' option, "
            "and you specified where the uncompressed sequences to use are "
            "('-d sequence_database_path'") in err
    # # Check output files
    summary =  os.path.join(GENEPATH, "assembly_summary-123.txt")
    assert not os.path.isfile(summary)
    ngd_outdir = os.path.join(GENEPATH, "genbank", "bacteria")
    assert not os.path.isdir(ngd_outdir)
    # Check logfiles are here
    log_files = glob.glob(os.path.join(GENEPATH, "*log*"))
    assert len(log_files) == 3
    # Check tmp files folder created, but empty asnothing is downloaded
    assert len(os.listdir(os.path.join(GENEPATH, "tmp_files"))) == 0
    # Check Database_init folder created, with at list 4 ".fna" genomes
    assert not os.path.isdir(os.path.join(GENEPATH, "Database_init"))


def test_main_norefseq_nodefault_dbdir_but_refseq(capsys):
    """
    We run with option norefseq, but given db_dir does not exist.
    -> error message
    """
    NCBI_species_name = ""
    NCBI_species_taxid = "123"
    NCBI_taxid = ""
    NCBI_section = "genbank"
    levels = ""
    # Copy refseq/bacteria and content into outdirectory
    outdir = GENEPATH
    tmp_dir = ""
    threads = 1
    norefseq = True
    orig_dbdir = os.path.join(GEN_PATH, "refseq")
    refseq_db_dir = os.path.join(GENEPATH, "genbank")
    shutil.copytree(orig_dbdir, refseq_db_dir)
    db_dir = ""
    only_mash = False
    l90 = 100
    nbcont = 999
    cutn = 0
    min_dist = 1e-4
    max_dist = 0.06
    verbose = 2
    quiet = False
    info_file = ""
    out_info_file = os.path.join(outdir, f"LSTINFO-123-filtered-0.0001_0.06.txt")
    assert prepare.main("cmd", NCBI_species_name, NCBI_species_taxid, NCBI_taxid, levels, 
                        NCBI_section, outdir, tmp_dir, threads,
                        norefseq, db_dir, only_mash, info_file, l90, nbcont, cutn, min_dist,
                        max_dist, verbose, quiet) == out_info_file
    out, err = capsys.readouterr()
    assert ("You asked to skip genbank downloads") in err
    assert ("Database folder test/data/prepare/generated_by_func-tests/"
            "Database_init supposed "
            "to contain fasta sequences does not exist. We will check if the download folder "
            "(with compressed sequences) exists.") in err
    assert ("Uncompressing genome files") in out
    assert ("Total number of genomes for 123: 3") in out
    assert ("Computing pairwise distances between all genomes") in out
    assert ("Final number of genomes in dataset: 1") in out
    # Check output files
    # Check that the NCBI_genome_download output directory exists
    ngd_outdir = os.path.join(GENEPATH, "genbank", "bacteria")
    # And that it contains folders
    assert os.path.isdir(ngd_outdir)
    assert len(os.listdir(ngd_outdir)) == 3
    # Check logfiles are here
    log_files = glob.glob(os.path.join(GENEPATH, "*log*"))
    assert len(log_files) == 3
    # Check tmp files folder created, but empty as we do not split
    tmp_folder = os.listdir(os.path.join(GENEPATH, "tmp_files"))
    assert len(tmp_folder) == 0
    # Check Database_init folder created, with the 3 ".fna" genomes
    fna_files = glob.glob(os.path.join(GENEPATH, "Database_init", "*.fna"))
    assert len(fna_files) == 3


def test_main_norefseq_defaultdbdir(capsys):
    """
    We run with option norefseq, but given db_dir does not exist.
    -> error message
    """
    NCBI_species_name = ""
    NCBI_species_taxid = ""
    NCBI_taxid = ""
    NCBI_section = "refseq"
    levels = ""
    # Copy refseq/bacteria and content into outdirectory
    outdir = GENEPATH
    tmp_dir = ""
    threads = 1
    norefseq = True
    orig_dbdir = os.path.join(GEN_PATH, "genomes_comparison")
    refseq_db_dir = os.path.join(GENEPATH, "Database_init")
    shutil.copytree(orig_dbdir, refseq_db_dir)
    db_dir = ""
    only_mash = False
    l90 = 100
    nbcont = 999
    cutn = 0
    min_dist = 1e-4
    max_dist = 0.06
    verbose = 2
    quiet = False
    info_file = ""
    out_info_file = os.path.join(outdir, "LSTINFO-NA-filtered-0.0001_0.06.txt")
    assert prepare.main("cmd", NCBI_species_name, NCBI_species_taxid, NCBI_taxid, levels, 
                        NCBI_section, outdir, tmp_dir, threads,
                        norefseq, db_dir, only_mash, info_file, l90, nbcont, cutn, min_dist,
                        max_dist, verbose, quiet) == out_info_file
    out, err = capsys.readouterr()
    assert ("You asked to skip refseq downloads") in err
    assert ("Total number of genomes for NA: 5") in out
    assert ("Computing pairwise distances between all genomes") in out
    assert ("Final number of genomes in dataset: 1") in out
    # Check output files
    # Check that the NCBI_genome_download output directory exists
    ngd_outdir = os.path.join(GENEPATH, "refseq", "bacteria")
    assert not os.path.isdir(ngd_outdir)
    # Check logfiles are here
    log_files = glob.glob(os.path.join(GENEPATH, "*log*"))
    assert len(log_files) == 3
    # Check tmp files folder created, but empty as we do not split
    tmp_folder = os.listdir(os.path.join(GENEPATH, "tmp_files"))
    assert len(tmp_folder) == 0
    # Check Database_init folder created, with the 3 ".fna" genomes
    fna_files = glob.glob(os.path.join(GENEPATH, "Database_init", "*.fna"))
    assert len(fna_files) == 5


def test_main_norefseq_givendbdir(capsys):
    """
    We run with option norefseq, but given db_dir does not exist.
    -> error message
    """
    NCBI_species_name = ""
    NCBI_species_taxid = ""
    NCBI_taxid = ""
    NCBI_section = "refseq"
    levels = ""
    # Copy refseq/bacteria and content into outdirectory
    outdir = GENEPATH
    tmp_dir = ""
    threads = 1
    norefseq = True
    orig_dbdir = os.path.join(GEN_PATH, "genomes_comparison")
    refseq_db_dir = os.path.join(GENEPATH, "genomes_comparison")
    shutil.copytree(orig_dbdir, refseq_db_dir)
    db_dir = refseq_db_dir
    only_mash = False
    l90 = 100
    nbcont = 999
    cutn = 2
    min_dist = 1e-4
    max_dist = 0.06
    verbose = 2
    quiet = False
    info_file = ""
    out_info_file = os.path.join(outdir, "LSTINFO-NA-filtered-0.0001_0.06.txt")
    assert prepare.main("cmd", NCBI_species_name, NCBI_species_taxid, NCBI_taxid, levels, 
                        NCBI_section, outdir, tmp_dir, threads,
                        norefseq, db_dir, only_mash, info_file, l90, nbcont, cutn, min_dist,
                        max_dist, verbose, quiet) == out_info_file
    out, err = capsys.readouterr()
    assert ("You asked to skip refseq downloads") in err
    assert ("Total number of genomes for NA: 5") in out
    assert ("Computing pairwise distances between all genomes") in out
    assert ("Final number of genomes in dataset: 1") in out
    # Check output files
    # Check that the NCBI_genome_download output directory exists
    ngd_outdir = os.path.join(GENEPATH, "refseq", "bacteria")
    assert not os.path.isdir(ngd_outdir)
    # Check logfiles are here
    log_files = glob.glob(os.path.join(GENEPATH, "*log*"))
    assert len(log_files) == 3
    # Check tmp files folder created, but empty as we do not split
    tmp_files = glob.glob(os.path.join(GENEPATH, "tmp_files", "*.fna_prepare-split2N.fna"))
    assert len(tmp_files) == 5


def test_only_mash(capsys):
    """
    Running only mash step (giving genomes and corresponding LSTINFO file)
    """
    NCBI_species_name = ""
    NCBI_species_taxid = ""
    NCBI_taxid = ""
    NCBI_section = "refseq"
    levels = ""
    outdir = GENEPATH
    tmp_dir = ""
    threads = 1
    norefseq = False
    db_dir = ""
    only_mash = True
    info_file = os.path.join(TEST_DIR, "test_lstinfo_onlymash.lst")
    l90 = 100
    nbcont = 999
    cutn = 5
    min_dist = 1e-4
    max_dist = 0.06
    verbose = 1
    quiet = False
    out_info_file = os.path.join(outdir, "LSTINFO-NA-filtered-0.0001_0.06.txt")
    assert prepare.main("cmd", NCBI_species_name, NCBI_species_taxid, NCBI_taxid, levels, 
                        NCBI_section, outdir, tmp_dir, threads,
                        norefseq, db_dir, only_mash, info_file, l90, nbcont, cutn, min_dist,
                        max_dist, verbose, quiet) == out_info_file
    out, err = capsys.readouterr()
    assert ("You asked to run only mash steps") in err
    assert ("You want to run only mash steps. Getting information from "
            "test/data/prepare/test_files/test_lstinfo_onlymash.lst") in out
    assert ("Found 5 genomes in total") in out
    assert ("Computing pairwise distances between all genomes") in out
    assert ("Sorting all 5 genomes by quality") in out
    assert ("Final number of genomes in dataset: 1") in out

    # Check output files
    assert len(os.listdir(os.path.join(outdir, "tmp_files"))) == 0
    # Check logfiles are here
    log_files = glob.glob(os.path.join(outdir, "*log*"))
    assert len(log_files) == 3
    # Check content of output lstinfo file
    out_lst = os.path.join(outdir, "LSTINFO-NA-filtered-0.0001_0.06.txt")
    exp_lst = os.path.join(DBDIR, "exp_files", "exp_lstinfo_run_only-mash.lst")
    assert tutil.compare_order_content(out_lst, exp_lst)


def test_only_mash_empty_lstinfo(capsys):
    """
    Running only mash step giving an empty lstinfo file -> error, no genome found
    """
    NCBI_species_name = ""
    NCBI_species_taxid = ""
    NCBI_taxid = ""
    NCBI_section = "refseq"
    levels = ""
    outdir = GENEPATH
    tmp_dir = ""
    threads = 1
    norefseq = False
    db_dir = ""
    only_mash = True
    # Create empty lstinfo file
    info_file = os.path.join(GENEPATH, "LSTINFO-empty.lst")
    open(info_file, "w").close()
    l90 = 100
    nbcont = 999
    cutn = 5
    min_dist = 1e-4
    max_dist = 0.06
    verbose = 1
    quiet = False
    with pytest.raises(SystemExit):
        prepare.main("cmd", NCBI_species_name, NCBI_species_taxid, NCBI_taxid, levels,
                     NCBI_section, outdir, tmp_dir, threads, norefseq,
                     db_dir, only_mash, info_file, l90, nbcont, cutn, min_dist, max_dist,
                     verbose, quiet)
    out, err = capsys.readouterr()
    assert ("You asked to run only mash steps") in err
    assert ("You want to run only mash steps. Getting information from "
            "test/data/prepare/generated_by_func-tests/LSTINFO-empty.lst") in out
    assert ("No genome listed in test/data/prepare/generated_by_func-tests/LSTINFO-empty.lst "
            "was found.") in err

    # Check output files
    assert len(os.listdir(os.path.join(outdir, "tmp_files"))) == 0
    # Check logfiles are here
    log_files = glob.glob(os.path.join(outdir, "*log*"))
    assert len(log_files) == 3
    # Check lstinfo file is still here and still empty
    assert os.stat(info_file).st_size == 0


def test_only_mash_no_lstinfo(capsys):
    """
    Running only mash step giving an info file which does not exist -> error missing infofile
    """
    NCBI_species_name = ""
    NCBI_species_taxid = ""
    NCBI_taxid = ""
    NCBI_section = "refseq"
    levels = ""
    outdir = GENEPATH
    tmp_dir = ""
    threads = 1
    norefseq = False
    db_dir = ""
    only_mash = True
    # Create empty lstinfo file
    info_file = "info_file.lst"
    l90 = 100
    nbcont = 999
    cutn = 5
    min_dist = 1e-4
    max_dist = 0.06
    verbose = 1
    quiet = False
    with pytest.raises(SystemExit):
        prepare.main("cmd", NCBI_species_name, NCBI_species_taxid, NCBI_taxid, levels,
                     NCBI_section, outdir, tmp_dir, threads, norefseq,
                     db_dir, only_mash, info_file, l90, nbcont, cutn, min_dist, max_dist,
                     verbose, quiet)
    out, err = capsys.readouterr()
    assert ("You asked to run only mash steps") in err
    assert ("Your info file info_file.lst does not exist. Please provide the  "
            "right name/path, or remove the '--mash-only option to rerun "
            "quality control.") in err

    # Check output files
    assert len(os.listdir(os.path.join(outdir, "tmp_files"))) == 0
    # Check logfiles are here
    log_files = glob.glob(os.path.join(outdir, "*log*"))
    assert len(log_files) == 3
    # Check that outdir contains only 4 elements: 3 logs + tmp_files repo
    files = os.listdir(outdir)
    files = [f for f in files if "fuse" not in f]
    assert len(files) == 4



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for the pan_to_pergenome submodule in align module
"""
import os
import shutil
import logging

import pytest

import PanACoTA.align_module.pan_to_pergenome as p2p
import test.test_unit.utilities_for_tests as tutil

ALPATH = os.path.join("test", "data", "align")
EXPPATH = os.path.join(ALPATH, "exp_files")
TESTPATH = os.path.join(ALPATH, "test_files")
GENEPATH = os.path.join(ALPATH, "generated_by_unit-tests")


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    os.mkdir(GENEPATH)
    print("setup")

    yield
    shutil.rmtree(GENEPATH)
    print("teardown")


ALL_PROTS = {"ESCO1": {"ESCO1_00001": '1',
                       "ESCO1_00002": '4'},
             "ESCO2": {"ESCO2_00001": '1',
                       "ESCO2_22": '2',
                       "ESCO2_456": '4',
                       "ESCO2_46": '3'},
             "ESCO3": {"ESCO3_1": '2',
                       "ESCO3_12": '1',
                       "ESCO3_4564": '3',
                       "ESCO3_00123": '4',
                       "ESCO3_8": '2'},
             "ESCO4": {"ESCO4_00001": '1',
                       "ESCO4_00002": '4',
                       "ESCO4_00003": '3',
                       "ESCO4_00004": '2',
                       "ESCO4_00006": '4'},
             "ESCO5": {"ESCO5_1": '1',
                       "ESCO5_2": '3',
                       "ESCO5_3": '2',
                       "ESCO5_4": '4',
                       "ESCO5_5": '2'},
             "ESCO6": {"ESCO6_1": '4',
                       "ESCO6_2": '3',
                       "ESCO6_3": '1'}}
FAM_GENOMES = {'1': ["ESCO1", "ESCO2", "ESCO3", "ESCO4", "ESCO5", "ESCO6"],
               '2': ["ESCO2", "ESCO3", "ESCO4", "ESCO5"],
               '3': ["ESCO2", "ESCO3", "ESCO4", "ESCO5", "ESCO6"],
               '4': ["ESCO1", "ESCO2", "ESCO3", "ESCO4", "ESCO5", "ESCO6"]}
SEVERAL = {'1': [],
           '2': ["ESCO3", "ESCO5"],
           '3': [],
           '4': ["ESCO4"]}
ALL_GENOMES = ["ESCO1", "ESCO2", "ESCO3", "ESCO4", "ESCO5", "ESCO6"]


def test_get_per_genome(caplog):
    """
    Test that when giving a persistent genome file and a list of genomes,
    it creates all expected files in output/Listdir
    """
    caplog.set_level(logging.DEBUG)
    pers = os.path.join("test", "data", "persgenome", "exp_files", "exp_pers-floor-mixed.txt")
    list_gen = os.path.join(TESTPATH, "listfile.txt")
    dname = "TEST-all-gembase"
    outdir = os.path.join(GENEPATH, "test_get_per_genome")
    all_genomes, aldir, listdir, fams = p2p.get_per_genome(pers, list_gen, dname, outdir)
    assert ("Reading PersGenome and constructing lists of missing genomes "
            "in each family") in caplog.text
    exp_al = os.path.join(outdir, "Align-TEST-all-gembase")
    exp_list = os.path.join(outdir, "List-TEST-all-gembase")
    assert exp_al == aldir
    assert os.path.isdir(aldir)
    assert exp_list == listdir
    assert os.path.isdir(listdir)
    exp_genomes = ["GEN4.1111.00001", "GENO.0817.00001", "GENO.1216.00002", "GENO.1216.00003"]
    assert all_genomes == exp_genomes
    exp_fams = ['1', '3', '5', '8', '10', '11', '12']
    assert set(fams) == set(exp_fams)


def test_prot_per_strain():
    """
    Test parser of persistent genome file
    """
    pers = os.path.join("test", "data", "persgenome", "exp_files", "exp_pers-floor-mixed.txt")
    all_prots, fams_genomes, several = p2p.proteins_per_strain(pers)
    exp_several = {'1': ["GENO.1216.00002"],
                   '3': [],
                   '5': [],
                   '8': [],
                   '10': [],
                   '11': [],
                   '12': ["GENO.1216.00003"]}
    assert several == exp_several
    exp_fams = {'1': ["GEN4.1111.00001", "GENO.0817.00001", "GENO.1216.00002", "GENO.1216.00003"],
                '3': ["GEN4.1111.00001", "GENO.0817.00001", "GENO.1216.00002", "GENO.1216.00003"],
                '5': ["GEN4.1111.00001", "GENO.0817.00001", "GENO.1216.00002", "GENO.1216.00003"],
                '8': ["GEN4.1111.00001", "GENO.0817.00001", "GENO.1216.00002"],
                '10': ["GEN4.1111.00001", "GENO.0817.00001", "GENO.1216.00002"],
                '11': ["GEN4.1111.00001", "GENO.0817.00001", "GENO.1216.00002"],
                '12': ["GEN4.1111.00001", "GENO.0817.00001", "GENO.1216.00002", "GENO.1216.00003"]}
    assert fams_genomes == exp_fams
    exp_prots = {"GEN4.1111.00001": {"GEN4.1111.00001.b0001_00001": '1',
                                     "GEN4.1111.00001.b0001_00009": '3',
                                     "GEN4.1111.00001.i0001_00002": '5',
                                     "GEN4.1111.00001.i0001_00007": '8',
                                     "GEN4.1111.00001.i0001_00004": '10',
                                     "GEN4.1111.00001.i0001_00005": '11',
                                     "GEN4.1111.00001.i0001_00008": '12'
                                     },
                 "GENO.0817.00001": {"GENO.0817.00001.b0001_00002": '1',
                                     "GENO.0817.00001.b0002_00011": '3',
                                     "GENO.0817.00001.b0002_00003": '5',
                                     "GENO.0817.00001.i0002_00009": '8',
                                     "GENO.0817.00001.i0002_00004": '10',
                                     "GENO.0817.00001.i0002_00005": '11',
                                     "GENO.0817.00001.i0002_00010": '12'
                                     },
                 "GENO.1216.00002": {"GENO.1216.00002.b0001_00001": '1',
                                     "GENO.1216.00002.i0001_00002": '1',
                                     "GENO.1216.00002.b0002_00010": '3',
                                     "GENO.1216.00002.i0001_00003": '5',
                                     "GENO.1216.00002.b0001_00008": '8',
                                     "GENO.1216.00002.i0001_00005": '10',
                                     "GENO.1216.00002.i0001_00006": '11',
                                     "GENO.1216.00002.b0002_00009": '12'
                                     },
                 "GENO.1216.00003": {"GENO.1216.00003.i0001_00003": '1',
                                     "GENO.1216.00003.i0001_01010": '3',
                                     "GENO.1216.00003.i0080_00010": '5',
                                     "GENO.1216.00003.i0001_00004": '12',
                                     "GENO.1216.00003.i0001_01000": '12'
                                     }
                 }
    assert all_prots == exp_prots


def test_prot_per_strain_member_bis(caplog):
    """
    Test parser of persistent genome file when a same member is in 2 different families
    """
    caplog.set_level(logging.DEBUG)
    pers = os.path.join(TESTPATH, "pers_genome_member-bis.txt")
    all_prots, fams_genomes, several = p2p.proteins_per_strain(pers)
    assert "problem: ESCO2_2 already exists, in family 5. Conflict with family 32" in caplog.text
    exp_several = {'1': [], '5': [], '12': [], '32': []}
    assert several == exp_several
    exp_fams = {'1': ["ESCO_1", "ESCO2", "ESCO3", "ESCO4"], '5': ["ESCO_1", "ESCO2", "ESCO4"],
                '12': ["ESCO_1", "ESCO2", "ESCO3"], '32': ["ESCO_1", "ESCO2", "ESCO3", "ESCO4"]}
    assert fams_genomes == exp_fams
    exp_prots = {"ESCO_1": {"ESCO_1_1": '1', "ESCO_1_2": '5', "ESCO_1_3": '12', "ESCO_1_4": '32'},
                 "ESCO2": {"ESCO2_1": '1', "ESCO2_2": '32', "ESCO2_3": '12'},
                 "ESCO3": {"ESCO3_1": '1', "ESCO3_3": '12', "ESCO3_4": '32'},
                 "ESCO4": {"ESCO4_1": '1', "ESCO4_3": '5', "ESCO4_4": '32'}}
    assert all_prots == exp_prots


def test_get_genomes():
    """
    Test parser of list of genomes
    """
    lstfile = os.path.join("test", "data", "annotate", "exp_files", "results_test_func-default",
                           "LSTINFO-list_genomes-func-test-default.lst")
    all_genomes = p2p.get_all_genomes(lstfile)
    assert all_genomes == ["ESCO.1015.00001", "ESCO.1116.00002", "GENO.1015.00001"]


def test_write_getentry():
    """
    Test that when giving a list of genomes with their persistent gene names,
    it creates all expected files.
    """
    listdir = os.path.join(GENEPATH, "Listdir")
    aldir = os.path.join(GENEPATH, "Aldir")
    # Create align folder
    os.makedirs(listdir)
    dname = "TEST6"
    p2p.write_getentry_files(ALL_PROTS, SEVERAL, listdir, aldir, dname, ALL_GENOMES)
    # Check creation and content of all files
    genfiles = [os.path.join(listdir, "{}-getEntry_gen_ESCO{}.txt".format(dname, num)) for num in
                range(1, 7)]
    expgens = [os.path.join(EXPPATH, "exp_getentry-gen-ESCO{}.txt".format(num)) for num in
               range(1, 7)]
    for fexp, fout in zip(expgens, genfiles):
        print(fexp, fout)
        assert tutil.compare_file_content(fexp, fout)
    prtfiles = [os.path.join(listdir, "{}-getEntry_prt_ESCO{}.txt".format(dname, num)) for num in
                range(1, 7)]
    expprts = [os.path.join(EXPPATH, "exp_getentry-prt-ESCO{}.txt".format(num)) for num in
               range(1, 7)]
    for fexp, fout in zip(expprts, prtfiles):
        assert tutil.compare_file_content(fexp, fout)


def test_write_getentry_error(caplog):
    """
    Test that when giving a list of genomes with their persistent gene names,
    but for 2 genomes, there is no persistent gene, it exists, with an error message
    """
    caplog.set_level(logging.DEBUG)
    all_prots = {"ESCO1": {"ESCO1_00001": '1',
                           "ESCO1_00002": '4'},
                 "ESCO2": {"ESCO2_00001": '1',
                           "ESCO2_22": '2',
                           "ESCO2_456": '4',
                           "ESCO2_46": '3'},
                 "ESCO3": {"ESCO3_1": '2',
                           "ESCO3_12": '1',
                           "ESCO3_4564": '3',
                           "ESCO3_00123": '4',
                           "ESCO3_8": '2'},
                 "ESCO6": {"ESCO6_1": '4',
                           "ESCO6_2": '3',
                           "ESCO6_3": '1'}}
    several = {'1': [],
               '2': ["ESCO3"],
               '3': [],
               '4': []}
    listdir = os.path.join(GENEPATH, "Listdir")
    aldir = os.path.join(GENEPATH, "Aldir")
    # Create align folder
    os.makedirs(listdir)
    dname = "TEST6"
    with pytest.raises(SystemExit):
        p2p.write_getentry_files(all_prots, several, listdir, aldir, dname, ALL_GENOMES)
    assert ("There is not any protein for genome ESCO4 in any family! The program will close, "
            "please fix this problem to be able to run the alignments") in caplog.text
    assert ("There is not any protein for genome ESCO5 in any family! The program will close, "
            "please fix this problem to be able to run the alignments") in caplog.text
    # Check creation and content of all files
    genfiles = [os.path.join(listdir, "{}-getEntry_gen_ESCO{}.txt".format(dname, num)) for num in
                list(range(1, 4)) + [6]]
    expgens = [os.path.join(EXPPATH, "exp_getentry-gen-ESCO{}.txt".format(num)) for num in
               list(range(1, 4)) + [6]]
    for fexp, fout in zip(expgens, genfiles):
        assert tutil.compare_file_content(fexp, fout)
    prtfiles = [os.path.join(listdir, "{}-getEntry_prt_ESCO{}.txt".format(dname, num)) for num in
                list(range(1, 4)) + [6]]
    expprts = [os.path.join(EXPPATH, "exp_getentry-prt-ESCO{}.txt".format(num)) for num in
               list(range(1, 4)) + [6]]
    for fexp, fout in zip(expprts, prtfiles):
        assert tutil.compare_file_content(fexp, fout)


def test_write_genome_prt_exists():
    """
    Test that when only prt file exists, it overwrites it and generates
    expected prt and gen files
    """
    listdir = os.path.join(GENEPATH, "Listdir")
    aldir = os.path.join(GENEPATH, "Aldir")
    # Create align folder
    os.makedirs(listdir)
    dname = "test_write_genome"
    strain = "ESCO4"
    members = ALL_PROTS[strain]

    # Create prt file
    fileprt = os.path.join(listdir, f"{dname}-getEntry_prt_ESCO4.txt")
    with open(fileprt, "w") as prtf:
        prtf.write("Wrong prt file\n")
    p2p.write_genome_file(listdir, aldir, dname, strain, members, SEVERAL)

    # Check creation of files and content
    expprt = os.path.join(EXPPATH, "exp_getentry-prt-ESCO4_write-prt.txt")
    assert tutil.compare_file_content(fileprt, expprt)
    filegen = os.path.join(listdir, f"{dname}-getEntry_gen_ESCO4.txt")
    expgen = os.path.join(EXPPATH, "exp_getentry-gen-ESCO4_write-prt.txt")
    assert tutil.compare_file_content(expgen, filegen)


def test_write_genome_gen_exists():
    """
    Test that when only gen file exists, it overwrites it and generates
    expected prt and gen files
    """
    listdir = os.path.join(GENEPATH, "Listdir")
    aldir = os.path.join(GENEPATH, "Aldir")
    # Create align folder
    os.makedirs(listdir)
    dname = "test_write_genome"
    strain = "ESCO4"
    member4 = ALL_PROTS[strain]
    # Create prt file
    filegen = os.path.join(listdir, f"{dname}-getEntry_gen_ESCO4.txt")
    with open(filegen, "w") as genf:
        genf.write("Wrong gen file\n")
    p2p.write_genome_file(listdir, aldir, dname, strain, member4, SEVERAL)

    # Check creation of files and content
    fileprt = os.path.join(listdir, f"{dname}-getEntry_prt_ESCO4.txt")
    expprt = os.path.join(EXPPATH, "exp_getentry-prt-ESCO4_write-prt.txt")
    assert tutil.compare_file_content(fileprt, expprt)
    expgen = os.path.join(EXPPATH, "exp_getentry-gen-ESCO4_write-prt.txt")
    assert tutil.compare_file_content(filegen, expgen)


def test_write_genome_gen_prt_exist(caplog):
    """
    Test that when gen and prt files already exist, it does not do anything.
    Those files will be used for next steps.
    """
    caplog.set_level(logging.DEBUG)
    listdir = os.path.join(GENEPATH, "Listdir")
    aldir = os.path.join(GENEPATH, "Aldir")
    # Create align folder
    os.makedirs(listdir)
    dname = "TEST6"
    strain = "ESCO4"
    member4 = ALL_PROTS[strain]
    # Create gen and prt files
    filegen = os.path.join(listdir, "{}-getEntry_gen_ESCO4.txt".format(dname))
    with open(filegen, "w") as genf:
        genf.write("Wrong gen file\n")
    fileprt = os.path.join(listdir, "{}-getEntry_prt_ESCO4.txt".format(dname))
    with open(fileprt, "w") as prtf:
        prtf.write("Wrong prt file\n")
    p2p.write_genome_file(listdir, aldir, dname, strain, member4, SEVERAL)

    # Check log
    assert ("For genome ESCO4, "
            "test/data/align/generated_by_unit-tests/Listdir/TEST6-getEntry_prt_ESCO4.txt and "
            "test/data/align/generated_by_unit-tests/Listdir/TEST6-getEntry_gen_ESCO4.txt "
            "already exist. The program will use them to extract "
            "proteins and genes. If you prefer to rewrite them, use option "
            "-F (or --force).".format(fileprt, filegen)) in caplog.text

    # Check content of prt and gen has not changed
    with open(fileprt, "r") as prtf:
        lines = prtf.readlines()
        assert lines == ["Wrong prt file\n"]
    with open(filegen, "r") as prtf:
        lines = prtf.readlines()
        assert lines == ["Wrong gen file\n"]


def test_write_genome():
    """
    Test that given a genome, it writes the list of its proteins
    and genes in expected files.
    """
    listdir = os.path.join(GENEPATH, "Listdir")
    aldir = os.path.join(GENEPATH, "Aldir")
    # Create align folder
    os.makedirs(listdir)
    dname = "test_write_genome"
    strain = "ESCO4"
    members = ALL_PROTS[strain]
    p2p.write_genome_file(listdir, aldir, dname, strain, members, SEVERAL)

    # Check creation of files and content
    fileprt = os.path.join(listdir, f"{dname}-getEntry_prt_ESCO4.txt")
    expprt = os.path.join(EXPPATH, "exp_getentry-prt-ESCO4_write-prt.txt")
    assert tutil.compare_file_content(fileprt, expprt)
    filegen = os.path.join(listdir, f"{dname}-getEntry_gen_ESCO4.txt")
    expgen = os.path.join(EXPPATH, "exp_getentry-gen-ESCO4_write-prt.txt")
    assert tutil.compare_file_content(filegen, expgen)


def test_write_missing():
    """
    Test that given families with genomes present, genomes with several numbers and
    list of all genomes, it returns, for each family, the genomes which will not
    be considered.
    """
    dname = "test_write_missing"
    p2p.write_missing_genomes(FAM_GENOMES, SEVERAL, ALL_GENOMES, GENEPATH, dname)

    exp_res = [None, [], ["ESCO1", "ESCO3", "ESCO5", "ESCO6"], ["ESCO1"], ["ESCO4"]]
    for num in range(1,5):
        miss_file = os.path.join(GENEPATH, f"{dname}-current.{num}.miss.lst")
        assert os.path.isfile(miss_file)
        assert tutil.compare_file_to_list(miss_file, exp_res[num])



# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
PanACoTA align submodule
"""



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Functions to:

- analyse a genome (cut at stretch of N if asked, calc L90, nb cont, size...)
- if genome cut by stretch of N, write the new sequence to new file
- rename genome contigs and write new sequence to new file


@author gem
April 2017
"""
import os
import re
import sys
import numpy as np
import logging
import progressbar

from PanACoTA import utils

logger = logging.getLogger("annotate.gseq_functions")

def analyse_all_genomes(genomes, dbpath, tmp_path, nbn, soft, logger, quiet=False):
    """

    Parameters
    ----------
    genomes : dict
        {genome: spegenus.date}
    dbpath : str
        path to folder containing genomes
    tmp_path : str
        path to put out files
    nbn : int
        minimum number of 'N' required to cut into a new contig
    soft : str
        soft used (prokka, prodigal, or None if called by prepare module)
    logger : logging.Logger
        logger object to write log information. Because this function can be called from
        prepare module, where sub logger name is different
    quiet : bool
        True if nothing must be written to stdout/stderr, False otherwise

    Returns
    -------
    dict
        {genome: [spegenus.date, orig_name, path_to_seq_to_annotate, size, nbcont, l90]}

    """
    cut = nbn > 0
    pat = None  ## To put pattern with which sequence must be cut
    if cut:
        pat = 'N' * nbn + "+"
    nbgen = len(genomes)
    bar = None
    curnum = None
    if cut:
        logger.info(("Cutting genomes at each time there are at least {} 'N' in a row, "
                     "and then, calculating genome size, number of contigs and L90.").format(nbn))
    else:
        logger.info("Calculating genome size, number of contigs, L90")
    if not quiet:
        # Create progressbar
        widgets = ['Analysis: ', progressbar.Bar(marker='█', left='', right=''),
                   ' ', progressbar.Counter(), "/{}".format(nbgen), ' (',
                   progressbar.Percentage(), ') - ', progressbar.Timer(), ' - ',
                   progressbar.ETA()
                   ]
        bar = progressbar.ProgressBar(widgets=widgets, max_value=nbgen, term_width=79).start()
        curnum = 1
    toremove = []
    # Analyse genomes 1 by 1
    for genome, name in genomes.items():
        # If not quiet option, show progress bar
        if not quiet:
            bar.update(curnum)
            curnum += 1
        # analyse genome, and check everything went well.
        # exception if binary file
        try:
            res = analyse_genome(genome, dbpath, tmp_path, cut, pat, genomes, soft, logger=logger)
        except UnicodeDecodeError:
            logger.warning(f"'{genome}' does not seem to be a fasta file. It will be ignored.")
            res = False
        # Problem while analysing genome -> genome ignored
        if not res:
            toremove.append(genome)
    # If there are some genomes to remove (analysis failed), remove them from genomes dict.
    if toremove:
        for gen in toremove:
            del genomes[gen]
    if not genomes:
        logger.error(f"No genome was found in the database folder {dbpath}. See logfile "
                     "for more information.")
        sys.exit(1)
    if not quiet:
        bar.finish()
    return 0


def analyse_genome(genome, dbpath, tmp_path, cut, pat, genomes, soft, logger):
    """
    Analyse given genome:

    - if cut is asked:

        - cut its contigs at each time that 'pat' is seen
        - save cut genome in new file

    - calculate genome size, L90, nb contigs and save it into genomes

    Parameters
    ----------
    genome : str
        given genome to analyse
    dbpath : str
        path to the folder containing the given genome sequence
    tmp_path : str
        path to folder where output files must be saved.
    cut : bool
        True if contigs must be cut, False otherwise
    pat : str
        pattern on which contigs must be cut. ex: "NNNNN"
    genomes : dict
        {genome_file: [genome_name]} as input, and will be changed to\
        {genome_file: [genome_name, path, path_annotate, gsize, nbcont, L90]}
    soft : str
        soft used (prokka, prodigal, or None if called by prepare module)

    Returns
    -------
    bool
        True if genome analysis went well, False otherwise
        Modifies 'genomes' for the analysed genome: -> {genome_file: [genome_name, path,
        path_annotate, gsize, nbcont, L90]}
    """
    gpath, grespath = get_output_dir(soft, dbpath, tmp_path, genome, cut, pat)
    if not os.path.exists(gpath):
        logger.error(f"The file {gpath} does not exist")
        return False
    # Open original sequence file
    with open(gpath, "r") as genf:
        # If a new file must be created (sequences cut), open it
        gresf = None
        if grespath:
            gresf = open(grespath, "w")

        # Initialize variables
        cur_contig_name = "" # header text
        contig_sizes = {}  # {header text: size}
        cur_seq = "" # sequence
        num = 1 # Used to get unique contig names

        # Read each line of original sequence
        for line in genf:
            #### NEW CONTIG
            # Line corresponding to a new contig
            if line.startswith(">"):
                # If not first contig, write info to output file (if needed)
                if cur_seq != "":
                    num = format_contig(cut, pat, cur_seq, cur_contig_name, genome, contig_sizes,
                                        gresf, num, logger)
                    # If problem while formatting contig, return False -> genome ignored
                    if num == -1:
                        return False
                # Get contig name for next turn, and reset sequence
                cur_contig_name = line.strip()
                # Initialize for next contig
                cur_seq = ""
            # #### SEQUENCE LINE
            # If sequence line, keep it, all in upper case
            else:
                # Add this line without \n to sequence (1 sequence per line)
                cur_seq += line.strip().upper()

        # LAST CONTIG
        if cur_contig_name != "":
            num = format_contig(cut, pat, cur_seq, cur_contig_name, genome, contig_sizes, gresf,
                                num, logger)
            if num == -1:
                return False
    # GLOBAL INFORMATION
    nbcont = len(contig_sizes)
    gsize = sum(contig_sizes.values())
    if nbcont == 0 or gsize == 0:
        logger.warning(f"Your file {gpath} does not contain any gene. Please check that you "
                       "really gave a fasta sequence file")
        if grespath:
            gresf.close()
            os.remove(grespath)
        return False
    l90 = calc_l90(contig_sizes)
    # Everything ok for this genome -> complete its list of information in genomes dict
    # [spegenus.date] -> [spegenus.date, gpath, gpath_to_annotate, gsize, nbcont, L90]}
    if grespath:
        genomes[genome] += [gpath, grespath, gsize, nbcont, l90]
    else:
        genomes[genome] += [gpath, gpath, gsize, nbcont, l90]
    # If we wrote a new sequence file, close it
    if grespath:
        gresf.close()
    return True


def get_output_dir(soft, dbpath, tmp_path, genome, cut, pat):
    """
    Get output file to put sequence cut and/or sequence with shorter contigs (prokka)

    Parameters
    ----------
    soft : str
        soft used (prokka, prodigal, or None if called by prepare module)
    dbpath : str
        path to the folder containing the given genome sequence
    tmp_path : str
        path to folder where output files must be saved.
    genome : str
        genome name
    cut : bool
        True if contigs must be cut, False otherwise
    pat : str
        pattern on which contigs must be cut. ex: "NNNNN"

    Return
    ------
    grespath : str
        path to ouput file. None if no need to create new sequence file
    """
    # Path to sequence to analyze
    gpath = os.path.join(dbpath, genome)
    # genome file is in dbpath except if it was in several files in dbpath,
    # in which case it has been concatenated to a file in tmp_path
    if not os.path.isfile(gpath):
        gpath = os.path.join(tmp_path, genome)
    # New file create if needed. If not (prodigal and not cut), empty filename
    grespath = None
    # If user asks to cut at each 'pat', need to create a new sequence file,
    # whatever the annotation soft used
    if cut:
        new_file = genome + "_{}-split{}N.fna".format(soft, len(pat) - 1)
        grespath = os.path.join(tmp_path, new_file)
    # If no cutl, just keep original sequence, no need to create new file.
    # Just check that contigs have different names
    return gpath, grespath


def format_contig(cut, pat, cur_seq, cur_contig_name, genome, contig_sizes, gresf, num, logger):
    """
    Format given contig, and save to output file if needed

    - if cut: cut it and write each subsequence
    - write new contig just check that contig names are different

    Parameters
    ----------
    cut : bool
        True if contigs must be cut, False otherwise
    pat : str
        pattern on which contigs must be cut. ex: "NNNNN"
    cur_seq : str
        current sequence (aa)
    cur_contig_name : str
        name of current contig
    genome : str
        name of current genome
    cont_sizes : dict
        {contig_name : sequence length}
    gresf : io.TextIOWrappe
        open file to write new sequence. If we are annotating with prodigal and not cutting,
        there is no new sequence -> gref is None
    num : int
        current contig number
    logger : logging.Logger
        logger object to write log information

    Returns
    -------
    bool
        True if contig has been written without problem, False if any problem
    """
    # "CUT" if cut: need to cut at each 'pat' -> write new header + seq in new file
    if cut:
        # Cut sequence and write header + sequence to res file
        num = split_contig(pat, cur_seq, cur_contig_name, contig_sizes, gresf, num)
    # No cut -> no new file created, but check contig unique names
    else:
        if cur_contig_name in contig_sizes.keys():
            logger.error(f"In genome {genome}, '{cur_contig_name}' contig name is used for "
                         "several contigs. Please put different names for each contig. "
                         "This genome will be ignored.")
            return -1
        else:
            contig_sizes[cur_contig_name] = len(cur_seq)
    return num


def split_contig(pat, whole_seq, cur_contig_name, contig_sizes, gresf, num):
    """
    Save the contig read just before into dicts and write it to sequence file.
    Unique ID of contig must be in the first field of header, before the first space
    (required by prokka)

    Parameters
    ----------
    pat : str
        pattern to split a contig. None if we do not want to split
    whole_seq : str
        sequence of current contig, to save once split according to pat
    cur_contig_name : str
        name of current contig to save once split according to pat
    contig_sizes : dict
        {name: size} save cur_contig once split according to pat
    gresf : _io.TextIOWrapper
        file open in w mode to save the split sequence
    num : int
        current contig number.

    Returns
    -------
    int
        new contig number, after giving number(s) to the current contig
    """
    # split contig each time a stretch of at least nbn 'N' is found (pattern pat)
    if not pat:
        cont_parts = [whole_seq]
    else:
        cont_parts = re.split(pat, whole_seq)

    # save contig parts
    for seq in cont_parts:
        # Only save non empty contigs (with some patterns, it could arrive that
        # we get empty contigs, if 2 occurrences of the pattern are side by side).
        if len(seq) == 0:
            continue
        new_contig_name = ">{}_{}\n".format(num, cur_contig_name.split(">")[1])
        contig_sizes[new_contig_name] = len(seq)
        gresf.write(new_contig_name)
        gresf.write(seq + "\n")
        num += 1
    return num


def calc_l90(contig_sizes):
    """
    Calc L90 of a given genome

    Parameters
    ----------
    contig_sizes : dict
        {name: size}

    Returns
    -------
    None or int
        if L90 found, returns L90. Otherwise, returns nothing
    """
    gsize = sum(contig_sizes.values())
    sizes = [contig_sizes[cont] for cont in contig_sizes]
    cum_sizes = np.cumsum(sorted(sizes, reverse=True))
    lim = 0.9 * gsize
    for num, val in enumerate(cum_sizes):
        if val >= lim:
            return num + 1


def rename_all_genomes(genomes):
    """
    FUNCTION DIRECTLY CALLED FROM MAIN ANNOTATE MODULE (step 3)
    Sort kept genomes by L90 and then nb contigs.
    For each genome, assign a strain number, and rename all its contigs.

    Parameters
    ----------
    genomes : dict
        {genome: [name, path, path_to_seq, gsize, nbcont, L90]} as input, and will become\
        {genome: [gembase_name, path, path_to_seq, gsize, nbcont, L90]} at the end

    Return
    ------
        change 1st field of genomes dict. name -> gembase_name (with strain number)

    """
    logger.info(f"Renaming kept genomes according to their quality ({len(genomes)} genomes)")
    # Keep first genome name to give to prodigal for training
    first_gname = None
    # Keep previous genome name (ESCO.0109 -> ESCO)
    last_name = ""
    # Keep last strain number
    last_strain = 0
    # "SAEN.1015.{}".format(str(last_strain).zfill(5))
    # Sort genomes by species, L90 and nb_contigs
    for genome, [name, _, _, _, _, _] in sorted(genomes.items(),
                                                key=utils.sort_genomes_byname_l90_nbcont):
        if not first_gname:
            first_gname = genome
        # first genome, or new strain name (ex: ESCO vs EXPL)
        # -> keep this new name, and add 1 to next strain number
        if last_name != name.split(".")[0]:
            last_strain = 1
            last_name = name.split(".")[0]
        # same strain name
        # -> write this new sequence, and go to next one (strain += 1)
        else:
            last_strain += 1
        # Write information to "genomes" dict.
        gembase_name = ".".join([name, str(last_strain).zfill(5)])
        genomes[genome][0] = gembase_name
    return first_gname


def plot_distributions(genomes, res_path, listfile_base, l90, nbconts):
    """
    FUNCTION DIRECTLY CALLED FROM MAIN ANNOTATE MODULE (step2)
    Plot distributions of L90 and nbcontig values.

    Parameters
    ----------
    genomes : dict
        {genome: [name, orig_path, to_annotate_path, size, nbcont, l90]}
    res_path : str
        path to put all output files
    listfile_base : str
        name of list file
    l90 : int
        L90 threshold
    nbconts : int
        nb contigs threshold

    Returns
    -------
    (l90_vals, nbcont_vals, dist1, dist2) :

    - l90_vals : list of l90 values for all genomes
    - nbcont_vals : list of nbcontigs for all genomes
    - dist1 : matplotlib figure of distribution of L90 values
    - dist2 : matplotlib figure of distribution of nb contigs values

    """
    logger.info("Generating distribution of L90 and #contigs graphs.")
    l90_vals = [val for _, (_, _, _, _, _, val) in genomes.items()]
    outl90 = os.path.join(res_path, "QC_L90-" + listfile_base + ".png")
    nbcont_vals = [val for _, (_, _, _, _, val, _) in genomes.items()]
    outnbcont = os.path.join(res_path, "QC_nb-contigs-" + listfile_base + ".png")
    dist1 = utils.plot_distr(l90_vals, l90, "L90 distribution for all genomes",
                             "max L90 =", logger)
    dist2 = utils.plot_distr(nbcont_vals, nbconts,
                             "Distribution of number of contigs among all genomes",
                             "max #contigs =", logger)
    dist1.savefig(outl90)
    dist2.savefig(outnbcont)
    return l90_vals, nbcont_vals, dist1, dist2



#!/usr/bin/env python3
# coding: utf-8

"""
Functional tests for the parser of align subcommand
"""
import argparse
import pytest

from PanACoTA.subcommands import align


def test_parser_noarg(capsys):
    """
    Test that when the script is called without any argument, an error message appears,
    indicating the required arguments.
    """
    parser = argparse.ArgumentParser(description="Align families", add_help=False)
    align.build_parser(parser)
    with pytest.raises(SystemExit):
        align.parse(parser, "".split())
    _, err = capsys.readouterr()
    assert "usage: " in err
    assert "-c COREPERS -l LIST_GENOMES -n DATASET_NAME -d DBPATH" in err
    assert "-o OUTDIR" in err
    assert "[--threads THREADS] [-F] [-v] [-q] [-h]" in err
    assert "[-h]" in err
    assert "the following arguments are required: -c, -l, -n, -d, -o" in err


def test_parser_thread_notint(capsys):
    """
    Test that when the number of threads given is not an int, it returns an error message
    """
    parser = argparse.ArgumentParser(description="Align families", add_help=False)
    align.build_parser(parser)
    with pytest.raises(SystemExit):
        align.parse(parser, "-l listgenome -n dname -d dbpath -o outdir --threads 1.5".split())
    _, err = capsys.readouterr()
    assert "argument --threads threads: invalid int value: 1.5" in err


def test_parser_thread_toomany(capsys):
    """
    Test that when the number of threads given is higher than the total number of threads,
    it returns the expected error message
    """
    import multiprocessing
    nb = multiprocessing.cpu_count()
    parser = argparse.ArgumentParser(description="Align families", add_help=False)
    align.build_parser(parser)
    with pytest.raises(SystemExit):
        align.parse(parser,
                    "-l listgenome -n dname -d dbpath -o outdir "
                    "--threads {}".format(nb + 3).split())
    _, err = capsys.readouterr()
    assert ("You have {} threads on your computer, you cannot ask for more: "
            "invalid value: {}".format(nb, nb+3)) in err


def test_parser_thread_neg(capsys):
    """
    Test that when the number of threads given is a negative number, it returns the expected
    error message.
    """
    parser = argparse.ArgumentParser(description="Align families", add_help=False)
    align.build_parser(parser)
    with pytest.raises(SystemExit):
        align.parse(parser, "-l listgenome -n dname -d dbpath -o outdir --threads -5".split())
    _, err = capsys.readouterr()
    assert ("Please provide a positive number of threads (or 0 for all threads): "
            "Invalid value: -5") in err


def test_parser_default():
    """
    Test that when run with the minimum required arguments, all default values are
    as expected.
    """
    parser = argparse.ArgumentParser(description="Align families", add_help=False)
    align.build_parser(parser)
    options = align.parse(parser, "-c cp -l listgenome -n dname -d dbpath -o outdir".split())
    assert options.corepers == "cp"
    assert options.list_genomes == "listgenome"
    assert options.dataset_name == "dname"
    assert options.dbpath == "dbpath"
    assert options.outdir == "outdir"
    assert options.threads == 1
    assert options.force is False
    assert options.verbose == 0
    assert options.quiet is False


def test_parser_allthreads():
    """
    Test that when run with 0 for --threads option, it returns the total number of threads in
    computer
    """
    import multiprocessing
    nb = multiprocessing.cpu_count()
    parser = argparse.ArgumentParser(description="Align families", add_help=False)
    align.build_parser(parser)
    options = align.parse(parser, "-c cp -l listgenome -n dname -d dbpath -o outdir "
                                  "--threads 0".split())
    assert options.corepers == "cp"
    assert options.list_genomes == "listgenome"
    assert options.dataset_name == "dname"
    assert options.dbpath == "dbpath"
    assert options.outdir == "outdir"
    assert options.threads == nb
    assert options.force is False
    assert options.verbose == 0
    assert options.quiet is False


def test_parser_all_ok():
    """
    Test that when all arguments given are ok, it builds expected parser
    """
    parser = argparse.ArgumentParser(description="Align families", add_help=False)
    align.build_parser(parser)
    options = align.parse(parser, "-c cp -l listgenome -n dname -d dbpath -o outdir "
                                  "--threads 1".split())
    assert options.corepers == "cp"
    assert options.list_genomes == "listgenome"
    assert options.dataset_name == "dname"
    assert options.dbpath == "dbpath"
    assert options.outdir == "outdir"
    assert options.threads == 1
    assert options.force is False
    assert options.verbose == 0
    assert options.quiet is False



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for annotate/annotation_functions.py
"""

import pytest
import os
import logging
import shutil

import test.test_unit.utilities_for_tests as tutil
import PanACoTA.utils as utils
import PanACoTA.annotate_module.annotation_functions as afunc


# Define variables used by several tests
DBDIR = os.path.join("test", "data", "annotate")
GEN_PATH = os.path.join(DBDIR, "genomes")
TEST_DIR = os.path.join(DBDIR, 'test_files')
GENEPATH = os.path.join(DBDIR, "generated_by_unit-tests")
LOGFILE_BASE = os.path.join(GENEPATH, "logfile.log")
LOGFILES = [LOGFILE_BASE + ext for ext in [".log", ".log.debug", ".log.details", ".log.err"]]


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    if os.path.isdir(GENEPATH):
        content = os.listdir(GENEPATH)
        for f in content:
            assert f.startswith(".fuse")
    else:
        os.mkdir(GENEPATH)
    print("setup")

    yield

    for f in LOGFILES:
        if os.path.exists(f):
            os.remove(f)
    shutil.rmtree(GENEPATH, ignore_errors=True)
    print("teardown")


# Create a logger with sublogger inside
def my_logger(name):
    """
    logger given to function called by a subprocess
    """
    import multiprocessing
    m = multiprocessing.Manager()
    q = m.Queue()
    qh = logging.handlers.QueueHandler(q)
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    root.handlers = []
    logging.addLevelName(utils.detail_lvl(), "DETAIL")
    root.addHandler(qh)
    return q, logging.getLogger(name)


def test_count_headers():
    """
    Count how many sequences there are in the given multi-fasta file
    """
    seqfile = os.path.join(GEN_PATH, "genome4.fasta")
    nb = afunc.count_headers(seqfile)
    assert nb == 5


def test_count_tbl():
    """
    Count the different features found in the tbl file, and return
    nbcont, nbCDS, nbGene, nbCRISPR
    """
    tblfile = os.path.join(TEST_DIR, "original_name.fna-prokkaRes", "prokka_out_for_test.tbl")
    ncont, ncds, ngene = afunc.count_tbl(tblfile)
    assert ncont == 6
    assert ncds == 14
    assert ngene == 16


def test_run_all_prodigal():
    """
    Check that there is no problem when running prodigal on all genomes
    Start and end are not necessarily in the same order (ex: start1, start2, end2, end1)
    """
    logger = my_logger("test_run_all_parallel_more_threads")
    utils.init_logger(LOGFILE_BASE, 0, 'test_run_all_parallel_more_threads')
    # genomes = {genome: [name, gpath, annot_path, size, nbcont, l90]}
    genome1 = "H299_H561.fasta"
    gpath1 = os.path.join(GEN_PATH, genome1)
    genome2 = "A_H738.fasta"
    gpath2 = os.path.join(GEN_PATH, genome2)
    genomes = {genome1: ["test_runall_1by1_1", gpath1, gpath1, 12656, 3, 0],
               genome2: ["test_runall_1by1_2", gpath2, gpath2, 456464645, 1, 465]}
    threads = 8
    force = False
    trn_gname = genome2
    final = afunc.run_annotation_all(genomes, threads, force, GENEPATH, trn_gname,
                                     prodigal_only=True, quiet=True)
    assert final[genome1]
    assert final[genome2]
    q = logger[0]
    assert q.qsize() == 10
    assert q.get().message == "Annotating all genomes with prodigal"
    assert q.get().message == "Prodigal will train using test/data/annotate/genomes/A_H738.fasta"
    assert q.get().message == ("prodigal command: prodigal -i "
                               "test/data/annotate/genomes/A_H738.fasta -t "
                               "test/data/annotate/generated_by_unit-tests/A_H738.fasta.trn")
    assert q.get().message == "End training on test/data/annotate/genomes/A_H738.fasta"
    messages = []
    for i in range(6):
        a = q.get().message
        messages.append(a)
    message_start_annot1 = ("Start annotating test_runall_1by1_1 "
                            "(from test/data/annotate/genomes/H299_H561.fasta sequence) "
                            "with Prodigal")
    message_start_annot2 = ("Start annotating test_runall_1by1_2 "
                            "(from test/data/annotate/genomes/A_H738.fasta sequence) "
                            "with Prodigal")
    # Check that all messages exist. We cannot know in which order,
    # as 'genomes' is a dict, hence unordered, and as computation is done in parallel
    assert message_start_annot1 in messages
    assert message_start_annot2 in messages
    # Prodigal cmd
    message_cmd1 = ("Prodigal command: prodigal -i test/data/annotate/genomes/H299_H561.fasta "
                    "-d test/data/annotate/generated_by_unit-tests/H299_H561.fasta-prodigalRes/"
                    "test_runall_1by1_1.ffn -a test/data/annotate/generated_by_unit-tests/"
                    "H299_H561.fasta-prodigalRes/test_runall_1by1_1.faa -f gff "
                    "-o test/data/annotate/generated_by_unit-tests/"
                    "H299_H561.fasta-prodigalRes/test_runall_1by1_1.gff -t "
                    "test/data/annotate/generated_by_unit-tests/A_H738.fasta.trn -q")
    message_cmd2 = ("Prodigal command: prodigal -i test/data/annotate/genomes/A_H738.fasta "
                    "-d test/data/annotate/generated_by_unit-tests/A_H738.fasta-prodigalRes/"
                    "test_runall_1by1_2.ffn -a test/data/annotate/generated_by_unit-tests/"
                    "A_H738.fasta-prodigalRes/test_runall_1by1_2.faa -f gff "
                    "-o test/data/annotate/generated_by_unit-tests/A_H738.fasta-prodigalRes/"
                    "test_runall_1by1_2.gff -t "
                    "test/data/annotate/generated_by_unit-tests/A_H738.fasta.trn -q")
    assert message_cmd1 in messages
    assert message_cmd2 in messages
    message_end_annot1 = ("End annotating test_runall_1by1_1 (from test/data/annotate/genomes/"
                          "H299_H561.fasta)")
    message_end_annot2 = ("End annotating test_runall_1by1_2 (from test/data/annotate/genomes/"
                          "A_H738.fasta)")
    assert message_end_annot1 in messages
    assert message_end_annot2 in messages


def test_run_all_prodigal_small():
    """
    Check that there is no problem when running prodigal with --small option on all genomes
    Start and end are not necessarily in the same order (ex: start1, start2, end2, end1)
    """
    logger = my_logger("test_run_all_prodigal_small")
    # genomes = {genome: [name, gpath, annot_path, size, nbcont, l90]}
    genome1 = "H299_H561.fasta"
    gpath1 = os.path.join(GEN_PATH, genome1)
    genome2 = "A_H738.fasta"
    gpath2 = os.path.join(GEN_PATH, genome2)
    genomes = {genome1: ["test_runall_1by1_1", gpath1, gpath1, 12656, 3, 0],
               genome2: ["test_runall_1by1_2", gpath2, gpath2, 456464645, 1, 465]}
    threads = 8
    force = False
    trn_gname = genome2
    final = afunc.run_annotation_all(genomes, threads, force, GENEPATH, trn_gname,
                                     prodigal_only=True, quiet=True, small=True)
    assert final[genome1]
    assert final[genome2]
    q = logger[0]
    # assert q.qsize() == 10
    assert q.get().message == "Annotating all genomes with prodigal"
    # assert q.get().message == "Prodigal will train using test/data/annotate/genomes/A_H738.fasta"
    # assert q.get().message == ("prodigal command: prodigal -i "
    #                            "test/data/annotate/genomes/A_H738.fasta -t "
    #                            "test/data/annotate/generated_by_unit-tests/A_H738.fasta.trn")
    # assert q.get().message == "End training on test/data/annotate/genomes/A_H738.fasta"
    messages = []
    for i in range(6):
        a = q.get().message
        messages.append(a)
    message_start_annot1 = ("Start annotating test_runall_1by1_1 "
                            "(from test/data/annotate/genomes/H299_H561.fasta sequence) "
                            "with Prodigal")
    message_start_annot2 = ("Start annotating test_runall_1by1_2 "
                            "(from test/data/annotate/genomes/A_H738.fasta sequence) "
                            "with Prodigal")
    # Check that all messages exist. We cannot know in which order,
    # as 'genomes' is a dict, hence unordered, and as computation is done in parallel
    assert message_start_annot1 in messages
    assert message_start_annot2 in messages
    # Prodigal cmd
    message_cmd1 = ("Prodigal command: prodigal -i test/data/annotate/genomes/H299_H561.fasta "
                    "-d test/data/annotate/generated_by_unit-tests/H299_H561.fasta-prodigalRes/"
                    "test_runall_1by1_1.ffn -a test/data/annotate/generated_by_unit-tests/"
                    "H299_H561.fasta-prodigalRes/test_runall_1by1_1.faa -f gff "
                    "-o test/data/annotate/generated_by_unit-tests/"
                    "H299_H561.fasta-prodigalRes/test_runall_1by1_1.gff -p meta -q")
    message_cmd2 = ("Prodigal command: prodigal -i test/data/annotate/genomes/A_H738.fasta "
                    "-d test/data/annotate/generated_by_unit-tests/A_H738.fasta-prodigalRes/"
                    "test_runall_1by1_2.ffn -a test/data/annotate/generated_by_unit-tests/"
                    "A_H738.fasta-prodigalRes/test_runall_1by1_2.faa -f gff "
                    "-o test/data/annotate/generated_by_unit-tests/A_H738.fasta-prodigalRes/"
                    "test_runall_1by1_2.gff -p meta -q")
    assert message_cmd1 in messages
    assert message_cmd2 in messages
    message_end_annot1 = ("End annotating test_runall_1by1_1 (from test/data/annotate/genomes/"
                          "H299_H561.fasta)")
    message_end_annot2 = ("End annotating test_runall_1by1_2 (from test/data/annotate/genomes/"
                          "A_H738.fasta)")
    assert message_end_annot1 in messages
    assert message_end_annot2 in messages


def test_run_all_prodigal_error_train():
    """
    Check that when we want to train on a genome but it fails, it returns False for all genomes
    Here, it fails because genome to train on is too small
    """
    logger = my_logger("test_run_all_parallel_more_threads")
    utils.init_logger(LOGFILE_BASE, 0, 'test_run_all_parallel_more_threads')
    # genomes = {genome: [name, gpath, annot_path, size, nbcont, l90]}
    genome1 = "H299_H561.fasta"
    gpath1 = os.path.join(GEN_PATH, genome1)
    genome2 = "A_H738.fasta"
    gpath2 = os.path.join(GEN_PATH, genome2)
    genomes = {genome1: ["test_runall_1by1_1", gpath1, gpath1, 12656, 3, 0],
               genome2: ["test_runall_1by1_2", gpath2, gpath2, 456464645, 1, 465]}
    threads = 8
    force = False
    trn_gname = genome1
    final = afunc.run_annotation_all(genomes, threads, force, GENEPATH, trn_gname,
                                     prodigal_only=True, quiet=True)
    assert not final[genome1]
    assert not final[genome2]
    q = logger[0]
    assert q.qsize() == 4
    assert q.get().message == "Annotating all genomes with prodigal"
    assert q.get().message == ("Prodigal will train using "
                               "test/data/annotate/genomes/H299_H561.fasta")
    assert q.get().message == ("prodigal command: prodigal -i "
                               "test/data/annotate/genomes/H299_H561.fasta -t "
                               "test/data/annotate/generated_by_unit-tests/H299_H561.fasta.trn")
    assert q.get().message == ("Error while trying to train prodigal on H299_H561.fasta. See "
                               "test/data/annotate/generated_by_unit-tests/"
                               "H299_H561.fasta.trn-prodigal-train.log.err.")


def test_run_all_prodigal_train_exists_error():
    """
    Check that when we give a wrong training file, it does not train again, but
    fails while annotating
    """
    logger = my_logger("test_run_prodigal_train_exist_error")
    utils.init_logger(LOGFILE_BASE, 0, 'test_run_prodigal_train_exist_error')
    # genomes = {genome: [name, gpath, annot_path, size, nbcont, l90]}
    genome1 = "toto.fasta"
    gpath1 = os.path.join(GEN_PATH, genome1)
    genome2 = "A_H738.fasta"
    gpath2 = os.path.join(GEN_PATH, genome2)
    genomes = {genome1: ["test_runall_1by1_1", gpath1, gpath1, 12656, 3, 0],
               genome2: ["test_runall_1by1_2", gpath2, gpath2, 456464645, 1, 465]}
    threads = 8
    force = False
    trn_gname = genome1
    # Create empty trn file
    trn_file = os.path.join(GENEPATH, "toto.fasta.trn")
    open(trn_file, "w").close()
    # Run annotation all
    final = afunc.run_annotation_all(genomes, threads, force, GENEPATH, trn_gname,
                                     prodigal_only=True, quiet=False)
    assert not final[genome1]
    assert not final[genome2]
    q = logger[0]
    assert q.qsize() == 9
    assert q.get().message == "Annotating all genomes with prodigal"
    assert q.get().message == ("Prodigal will train using "
                               "test/data/annotate/genomes/toto.fasta")
    assert q.get().message == ("A training file already exists (test/data/annotate/"
                               "generated_by_unit-tests/toto.fasta.trn). It will be used "
                               "to annotate all genomes.")
    # Check that all messages exist. We cannot know in which order,
    # as 'genomes' is a dict, hence unordered, and as computation is done in parallel
    messages = []
    for i in range(6):
        a = q.get().message
        messages.append(a)
    # Check start annotation messages
    message_start_annot1 = ("Start annotating test_runall_1by1_1 "
                            "(from test/data/annotate/genomes/toto.fasta sequence) "
                            "with Prodigal")
    message_start_annot2 = ("Start annotating test_runall_1by1_2 "
                            "(from test/data/annotate/genomes/A_H738.fasta sequence) "
                            "with Prodigal")
    assert message_start_annot1 in messages
    assert message_start_annot2 in messages
    # Prodigal cmd
    message_cmd1 = ("Prodigal command: prodigal -i test/data/annotate/genomes/toto.fasta "
                    "-d test/data/annotate/generated_by_unit-tests/toto.fasta-prodigalRes/"
                    "test_runall_1by1_1.ffn -a test/data/annotate/generated_by_unit-tests/"
                    "toto.fasta-prodigalRes/test_runall_1by1_1.faa -f gff "
                    "-o test/data/annotate/generated_by_unit-tests/"
                    "toto.fasta-prodigalRes/test_runall_1by1_1.gff -t "
                    "test/data/annotate/generated_by_unit-tests/toto.fasta.trn -q")
    message_cmd2 = ("Prodigal command: prodigal -i test/data/annotate/genomes/A_H738.fasta "
                    "-d test/data/annotate/generated_by_unit-tests/A_H738.fasta-prodigalRes/"
                    "test_runall_1by1_2.ffn -a test/data/annotate/generated_by_unit-tests/"
                    "A_H738.fasta-prodigalRes/test_runall_1by1_2.faa -f gff "
                    "-o test/data/annotate/generated_by_unit-tests/A_H738.fasta-prodigalRes/"
                    "test_runall_1by1_2.gff -t "
                    "test/data/annotate/generated_by_unit-tests/toto.fasta.trn -q")
    assert message_cmd1 in messages
    assert message_cmd2 in messages
    message_end_annot1 = ("Error while trying to run prodigal. See "
                          "test/data/annotate/generated_by_unit-tests/"
                          "toto.fasta-prodigal.log.err.")
    message_end_annot2 = ("Error while trying to run prodigal. See "
                          "test/data/annotate/generated_by_unit-tests/"
                          "A_H738.fasta-prodigal.log.err.")
    assert message_end_annot1 in messages
    assert message_end_annot2 in messages


def test_run_all_prodigal_train_exists_ok():
    """
    Check that when we want to train on a genome but it fails, it returns False for all genomes
    Here, it fails because genome to train on is too small
    """
    logger = my_logger("test_run_prodigal_train_exist_error")
    utils.init_logger(LOGFILE_BASE, 0, 'test_run_prodigal_train_exist_error')
    # genomes = {genome: [name, gpath, annot_path, size, nbcont, l90]}
    genome1 = "toto.fasta"
    gpath1 = os.path.join(GEN_PATH, genome1)
    genome2 = "A_H738.fasta"
    gpath2 = os.path.join(GEN_PATH, genome2)
    genomes = {genome1: ["test_runall_1by1_1", gpath1, gpath1, 12656, 3, 0],
               genome2: ["test_runall_1by1_2", gpath2, gpath2, 456464645, 1, 465]}
    threads = 8
    force = False
    trn_gname = genome1
    # Copy trn file to outdir, so that panacota detects that it already exists
    orig_trn_file = os.path.join(TEST_DIR, "A_H738-and-B2_A3_5.fna.trn")
    trn_file = os.path.join(GENEPATH, "toto.fasta.trn")
    shutil.copyfile(orig_trn_file, trn_file)
    # Run annotation all
    final = afunc.run_annotation_all(genomes, threads, force, GENEPATH, trn_gname,
                                     prodigal_only=True, quiet=False)
    assert not final[genome1]
    assert final[genome2]
    q = logger[0]
    assert q.qsize() == 9
    assert q.get().message == "Annotating all genomes with prodigal"
    assert q.get().message == ("Prodigal will train using "
                               "test/data/annotate/genomes/toto.fasta")
    assert q.get().message == ("A training file already exists (test/data/annotate/"
                               "generated_by_unit-tests/toto.fasta.trn). It will be used "
                               "to annotate all genomes.")
    # Check that all messages exist. We cannot know in which order,
    # as 'genomes' is a dict, hence unordered, and as computation is done in parallel
    messages = []
    for i in range(6):
        a = q.get().message
        messages.append(a)
    # Check start annotation messages
    message_start_annot1 = ("Start annotating test_runall_1by1_1 "
                            "(from test/data/annotate/genomes/toto.fasta sequence) "
                            "with Prodigal")
    message_start_annot2 = ("Start annotating test_runall_1by1_2 "
                            "(from test/data/annotate/genomes/A_H738.fasta sequence) "
                            "with Prodigal")
    assert message_start_annot1 in messages
    assert message_start_annot2 in messages
    # Prodigal cmd
    message_cmd1 = ("Prodigal command: prodigal -i test/data/annotate/genomes/toto.fasta "
                    "-d test/data/annotate/generated_by_unit-tests/toto.fasta-prodigalRes/"
                    "test_runall_1by1_1.ffn -a test/data/annotate/generated_by_unit-tests/"
                    "toto.fasta-prodigalRes/test_runall_1by1_1.faa -f gff "
                    "-o test/data/annotate/generated_by_unit-tests/"
                    "toto.fasta-prodigalRes/test_runall_1by1_1.gff -t "
                    "test/data/annotate/generated_by_unit-tests/toto.fasta.trn -q")
    message_cmd2 = ("Prodigal command: prodigal -i test/data/annotate/genomes/A_H738.fasta "
                    "-d test/data/annotate/generated_by_unit-tests/A_H738.fasta-prodigalRes/"
                    "test_runall_1by1_2.ffn -a test/data/annotate/generated_by_unit-tests/"
                    "A_H738.fasta-prodigalRes/test_runall_1by1_2.faa -f gff "
                    "-o test/data/annotate/generated_by_unit-tests/A_H738.fasta-prodigalRes/"
                    "test_runall_1by1_2.gff -t "
                    "test/data/annotate/generated_by_unit-tests/toto.fasta.trn -q")
    assert message_cmd1 in messages
    assert message_cmd2 in messages
    message_end_annot1 = ("Error while trying to run prodigal. See "
                          "test/data/annotate/generated_by_unit-tests/"
                          "toto.fasta-prodigal.log.err.")
    message_end_annot2 = ("End annotating test_runall_1by1_2 (from test/data/annotate/genomes/"
                          "A_H738.fasta)")
    assert message_end_annot1 in messages
    assert message_end_annot2 in messages


def test_run_all_prodigal_error_annotate():
    """
    Check running prodigal on 2 genomes:
    - prodigal train ok
    - running on genome1 ok
    - running on genome2 error
    """
    logger = my_logger("test_run_all_parallel_more_threads")
    utils.init_logger(LOGFILE_BASE, 0, 'test_run_all_parallel_more_threads')
    # genomes = {genome: [name, gpath, annot_path, size, nbcont, l90]}
    genome1 = "toto.fasta"
    gpath1 = os.path.join(GEN_PATH, genome1)
    genome2 = "A_H738.fasta"
    gpath2 = os.path.join(GEN_PATH, genome2)
    genomes = {genome1: ["test_runall_1by1_1", gpath1, gpath1, 12656, 3, 0],
               genome2: ["test_runall_1by1_2", gpath2, gpath2, 456464645, 1, 465]}
    threads = 8
    force = False
    trn_gname = genome2
    final = afunc.run_annotation_all(genomes, threads, force, GENEPATH, trn_gname,
                                     prodigal_only=True, quiet=False)
    assert not final[genome1]
    assert final[genome2]
    q = logger[0]
    assert q.qsize() == 10
    assert q.get().message == "Annotating all genomes with prodigal"
    assert q.get().message == "Prodigal will train using test/data/annotate/genomes/A_H738.fasta"
    assert q.get().message == ("prodigal command: prodigal -i "
                               "test/data/annotate/genomes/A_H738.fasta -t "
                               "test/data/annotate/generated_by_unit-tests/A_H738.fasta.trn")
    assert q.get().message == "End training on test/data/annotate/genomes/A_H738.fasta"
    messages = []
    for i in range(6):
        a = q.get().message
        messages.append(a)
    message_start_annot1 = ("Start annotating test_runall_1by1_1 "
                            "(from test/data/annotate/genomes/toto.fasta sequence) "
                            "with Prodigal")
    message_start_annot2 = ("Start annotating test_runall_1by1_2 "
                            "(from test/data/annotate/genomes/A_H738.fasta sequence) "
                            "with Prodigal")
    # Check that all messages exist. We cannot know in which order,
    # as 'genomes' is a dict, hence unordered, and as computation is done in parallel
    assert message_start_annot1 in messages
    assert message_start_annot2 in messages
    # Prodigal cmd
    message_cmd1 = ("Prodigal command: prodigal -i test/data/annotate/genomes/toto.fasta "
                    "-d test/data/annotate/generated_by_unit-tests/toto.fasta-prodigalRes/"
                    "test_runall_1by1_1.ffn -a test/data/annotate/generated_by_unit-tests/"
                    "toto.fasta-prodigalRes/test_runall_1by1_1.faa -f gff "
                    "-o test/data/annotate/generated_by_unit-tests/"
                    "toto.fasta-prodigalRes/test_runall_1by1_1.gff -t "
                    "test/data/annotate/generated_by_unit-tests/A_H738.fasta.trn -q")
    message_cmd2 = ("Prodigal command: prodigal -i test/data/annotate/genomes/A_H738.fasta "
                    "-d test/data/annotate/generated_by_unit-tests/A_H738.fasta-prodigalRes/"
                    "test_runall_1by1_2.ffn -a test/data/annotate/generated_by_unit-tests/"
                    "A_H738.fasta-prodigalRes/test_runall_1by1_2.faa -f gff "
                    "-o test/data/annotate/generated_by_unit-tests/A_H738.fasta-prodigalRes/"
                    "test_runall_1by1_2.gff -t "
                    "test/data/annotate/generated_by_unit-tests/A_H738.fasta.trn -q")
    assert message_cmd1 in messages
    assert message_cmd2 in messages
    message_end_annot1 = ("Error while trying to run prodigal. See "
                          "test/data/annotate/generated_by_unit-tests/"
                          "toto.fasta-prodigal.log.err.")
    message_end_annot2 = ("End annotating test_runall_1by1_2 (from test/data/annotate/genomes/"
                          "A_H738.fasta)")
    assert message_end_annot1 in messages
    assert message_end_annot2 in messages


def test_run_all_prodigal_outexists_error():
    """
    trn file already exists, and output folder too. No force option. Output folder is empty
    -> error message while checking prodigal
    """
    logger = my_logger("test_run_all_parallel_more_threads")
    utils.init_logger(LOGFILE_BASE, 0, 'test_run_all_parallel_more_threads')
    # genomes = {genome: [name, gpath, annot_path, size, nbcont, l90]}
    genome1 = "toto.fasta"
    genome2 = "A_H738.fasta"
    genomes = {genome1: ["test_runall_1by1_1", genome1, genome1, 12656, 3, 0],
               genome2: ["test_runall_1by1_2", genome2, genome2, 456464645, 1, 465]}
    # Create prodigal result directories
    prodigaldir_g1 = os.path.join(GENEPATH, "A_H738.fasta-prodigalRes")
    prodigaldir_g2 = os.path.join(GENEPATH, "toto.fasta-prodigalRes")
    os.makedirs(prodigaldir_g1)
    os.makedirs(prodigaldir_g2)
    # Other parameters
    threads = 1
    force = False
    # Add existing training file
    orig_trn_file = os.path.join(TEST_DIR, "A_H738-and-B2_A3_5.fna.trn")
    trn_file = os.path.join(GENEPATH, "toto.fasta.trn")
    shutil.copyfile(orig_trn_file, trn_file)
    trn_gname = genome1
    final = afunc.run_annotation_all(genomes, threads, force, GENEPATH, trn_gname,
                                     prodigal_only=True, quiet=False)
    assert not final[genome1]
    assert not final[genome2]
    q = logger[0]
    assert q.qsize() == 15
    assert q.get().message == "Annotating all genomes with prodigal"
    assert q.get().message == "Prodigal will train using toto.fasta"
    assert q.get().message == ("A training file already exists (test/data/annotate/"
                               "generated_by_unit-tests/toto.fasta.trn). It will "
                               "be used to annotate all genomes.")
    messages = []
    for i in range(12):
        a = q.get().message
        messages.append(a)
    message_start_annot1 = ("Start annotating test_runall_1by1_1 "
                            "(from toto.fasta sequence) with Prodigal")
    # Check that all messages exist. We cannot know in which order,
    # as 'genomes' is a dict, hence unordered, and as computation is done in parallel
    assert message_start_annot1 in messages
    # Prodigal cmd
    message_exists1 = ("Prodigal results folder test/data/annotate/generated_by_unit-tests/"
                    "toto.fasta-prodigalRes already exists.")
    message_errorfaa = ("test_runall_1by1_1 toto.fasta: no or several .faa file(s)")
    message_errorffn = ("test_runall_1by1_1 toto.fasta: no or several .ffn file(s)")
    message_errorgff = ("test_runall_1by1_1 toto.fasta: no or several .gff file(s)")
    message_error1 = ("Problems in the files contained in your already existing output dir "
                    "(test/data/annotate/generated_by_unit-tests/toto.fasta-prodigalRes). "
                    "Please check it, or remove it to re-annotate.")
    assert message_exists1 in messages
    assert message_errorfaa in messages
    assert message_errorffn in messages
    assert message_errorgff in messages
    assert message_error1 in messages
    message_start_annot2 = ("Start annotating test_runall_1by1_2 "
                            "(from A_H738.fasta sequence) with Prodigal")
    assert message_start_annot2 in messages
    message_error_annot2 = ("Problems in the files contained in your already existing output dir "
                          "(test/data/annotate/generated_by_unit-tests/A_H738.fasta-prodigalRes). "
                          "Please check it, or remove it to re-annotate.")
    assert message_error_annot2 in messages


def test_run_all_1by1_prokka():
    """
    Check that when running with 3 threads (not parallel), prokka runs as expected,
    and returns True for each genome
    -> Runs 1 by 1, with prokka using 3 cpus
    Start and end must be ordered: (start1, end1, start2, end2) or (start2, end2, start1, end1)
    """
    logger = my_logger("test_runall_1by1_1")
    utils.init_logger(LOGFILE_BASE, 0, 'test_run_all_1by1')
    # genomes = {genome: [name, gpath, size, nbcont, l90]}
    genome1 = "H299_H561.fasta"
    gpath1 = os.path.join(GEN_PATH, genome1)
    genome2 = "A_H738.fasta"
    gpath2 = os.path.join(GEN_PATH, genome2)
    genomes = {genome1: ["test_runall_1by1_1", gpath1, gpath1, 12656, 3, 0],
               genome2: ["test_runall_1by1_2", gpath2, gpath2, 456464645, 1, 465]}
    threads = 3
    force = False
    trn_file = "nofile.trn"
    annot_folder = os.path.join(GENEPATH, "annot-folder")
    os.makedirs(annot_folder)
    final = afunc.run_annotation_all(genomes, threads, force, annot_folder, trn_file)
    assert final[genome1]
    assert final[genome2]
    q = logger[0]
    assert q.qsize() == 7
    assert q.get().message == 'Annotating all genomes with prokka'
    # Messages for start and end annotation of the different genomes
    message_start_annot1 = ("Start annotating test_runall_1by1_1 test/data/annotate/genomes/"
                            "H299_H561.fasta")
    message_cmd1 = ("Prokka command: prokka --outdir test/data/annotate/generated_by_unit-tests/"
                    "annot-folder/H299_H561.fasta-prokkaRes --cpus 3")
    message_end_annot1 = ("End annotating test_runall_1by1_1 from test/data/annotate/genomes/"
                            "H299_H561.fasta.")
    message_start_annot2 = ("Start annotating test_runall_1by1_2 test/data/annotate/genomes/"
                            "A_H738.fasta")
    message_cmd2 = ("Prokka command: prokka --outdir test/data/annotate/generated_by_unit-tests/"
                    "annot-folder/A_H738.fasta-prokkaRes --cpus 3")
    message_end_annot2 = ("End annotating test_runall_1by1_2 from test/data/annotate/genomes/"
                            "A_H738.fasta.")
    qget = q.get().message
    # Check logs. Given that it is executed in parallel, we cannot know in which order messages
    # will appear
    assert qget == message_start_annot1 or message_start_annot2
    if qget == message_start_annot1:
        # Ending annotation of first genome (same genome as started because running 1by1)
        assert q.get().message.startswith(message_cmd1)
        assert q.get().message == message_end_annot1
    else:
        assert q.get().message.startswith(message_cmd2)
        assert q.get().message == message_end_annot2
    qget2 = q.get().message
    assert qget2 == message_start_annot1 or message_start_annot2
    if qget2 == message_start_annot2:
        # Ending annotation of first genome (same genome as started because running 1by1)
        assert q.get().message.startswith(message_cmd2)
        assert q.get().message == message_end_annot2
    else:
        assert q.get().message.startswith(message_cmd1)
        assert q.get().message == message_end_annot1


def test_run_all_prokka_parallel_less_threads():
    """
    Check that there is no problem when running with less threads than genomes (each genomes
    uses 2 threads)
    Genomes H299 and A_H738 should run well, but genomes genome* have problems (no CDS found),
    so check_prokka should return false.
    """
    logger = my_logger("test_run_all_parallel_more_threads")
    utils.init_logger(LOGFILE_BASE, 0, 'test_run_all_4threads')
    # genomes = {genome: [name, gpath, size, nbcont, l90]}
    gnames = ["H299_H561.fasta", "A_H738.fasta", "genome1.fasta", "genome2.fasta", "genome3.fasta"]
    gpaths = [os.path.join(GEN_PATH, name) for name in gnames]
    genomes = {gnames[0]: ["test_runall_1by1_1", gpaths[0], gpaths[0], 12656, 3, 1],
               gnames[1]: ["test_runall_1by1_2", gpaths[1], gpaths[1], 456464645, 1, 1],
               gnames[2]: ["test_runall_1by1_3", gpaths[2], gpaths[2], 456464645, 4, 1],
               gnames[3]: ["test_runall_1by1_4", gpaths[3], gpaths[3], 456464645, 3, 1],
               gnames[4]: ["test_runall_1by1_5", gpaths[4], gpaths[4], 456464645, 1, 1]
               }
    threads = 4
    force = False
    trn_file = "nofile.trn"
    final = afunc.run_annotation_all(genomes, threads, force, GENEPATH, trn_file)
    assert final[gnames[0]]
    assert final[gnames[1]]
    assert not final[gnames[2]]
    assert not final[gnames[3]]
    assert not final[gnames[4]]
    q = logger[0]
    # Check size of logs
    # -> starting log -> 1 log
    # -> for each genome ok (2 first ones): start annotate, prokka cmd, end annotate -> 6 logs
    # -> for each genome not ok (3 others):
    #           start annotate, prokka cmd, problem, end annotate -> 12 logs
    assert q.qsize() == 19
    assert q.get().message == "Annotating all genomes with prokka"
    # messages start annotation
    messages = []
    for i in range(18):
        a = q.get().message
        messages.append(a)
    message_start_annot1 = ("Start annotating test_runall_1by1_1 "
                            "from test/data/annotate/genomes/H299_H561.fasta "
                            "with Prokka")
    message_start_annot2 = ("Start annotating test_runall_1by1_2 "
                            "from test/data/annotate/genomes/A_H738.fasta "
                            "with Prokka")
    message_start_annot3 = ("Start annotating test_runall_1by1_4 "
                            "from test/data/annotate/genomes/genome2.fasta "
                            "with Prokka")
    # Check that all messages exist. We cannot know in which order,
    # as 'genomes' is a dict, hence unordered, and as computation is done in parallel
    assert message_start_annot1 in messages
    assert message_start_annot2 in messages
    assert message_start_annot3 in messages
    # messages Prokka cmd
    message_cmd1 = ("Prokka command: prokka --outdir test/data/annotate/generated_by_unit-tests/"
                    "H299_H561.fasta-prokkaRes --cpus 2 --prefix test_runall_1by1_1 "
                    "--centre prokka test/data/annotate/genomes/H299_H561.fasta")
    message_cmd2 = ("Prokka command: prokka --outdir test/data/annotate/generated_by_unit-tests/"
                    "A_H738.fasta-prokkaRes --cpus 2 --prefix test_runall_1by1_2 "
                    "--centre prokka test/data/annotate/genomes/A_H738.fasta")
    message_cmd3 = ("Prokka command: prokka --outdir test/data/annotate/generated_by_unit-tests/"
                    "genome1.fasta-prokkaRes --cpus 2 --prefix test_runall_1by1_3 "
                    "--centre prokka test/data/annotate/genomes/genome1.fasta")
    assert message_cmd1 in messages
    assert message_cmd2 in messages
    assert message_cmd3 in messages
    # Messages end annotation cmd
    message_end1 = ("End annotating test_runall_1by1_1 from "
                    "test/data/annotate/genomes/H299_H561.fasta.")
    message_end2 = ("End annotating test_runall_1by1_3 from "
                    "test/data/annotate/genomes/genome1.fasta.")
    message_end3 = ("End annotating test_runall_1by1_5 from "
                    "test/data/annotate/genomes/genome3.fasta.")
    assert message_end1 in messages
    assert message_end2 in messages
    assert message_end3 in messages
    # Messages error annotation cmd
    message_err1 = "test_runall_1by1_3 genome1.fasta: several .faa files"
    message_err2 = "test_runall_1by1_4 genome2.fasta: several .faa files"
    message_err3 = "test_runall_1by1_5 genome3.fasta: several .faa files"
    assert message_err1 in messages
    assert message_err2 in messages
    assert message_err3 in messages


def test_run_all_parallel_prokka_more_threads():
    """
    Check that there is no problem when running with more threads than genomes
    (6 threads and 2 genome: each genome uses 3 threads)
    Genomes H299 should run well but genome1.fasta should get an error
    """
    logger = my_logger("test_run_all_parallel_more_threads")
    utils.init_logger(LOGFILE_BASE, 0, 'test_run_all_4threads')
    # genomes = {genome: [name, gpath, size, nbcont, l90]}
    gnames = ["H299_H561.fasta", "genome1.fasta"]
    gpaths = [os.path.join(GEN_PATH, name) for name in gnames]
    genomes = {gnames[0]: ["test_runall_1by1_1", gpaths[0], gpaths[0], 12656, 3, 1],
               gnames[1]: ["test_runall_1by1_2", gpaths[1], gpaths[1], 456464645, 4, 1],
               }
    threads = 6
    force = False
    trn_file = "nofile.trn"
    final = afunc.run_annotation_all(genomes, threads, force, GENEPATH, trn_file)
    assert final[gnames[0]]
    assert not final[gnames[1]]
    q = logger[0]
    # Check size of logs
    # -> starting log -> 1 log
    # -> for genome ok : start annotate, prokka cmd, end annotate -> 3 logs
    # -> for genome not ok : start annotate, prokka cmd, problem, end annotate -> 4 logs
    assert q.qsize() == 8
    assert q.get().message == "Annotating all genomes with prokka"
    # messages start annotation
    messages = []
    for i in range(7):
        a = q.get().message
        messages.append(a)
    message_start_annot1 = ("Start annotating test_runall_1by1_1 "
                            "from test/data/annotate/genomes/H299_H561.fasta "
                            "with Prokka")
    message_start_annot2 = ("Start annotating test_runall_1by1_2 "
                            "from test/data/annotate/genomes/genome1.fasta "
                            "with Prokka")
    # Check that all messages exist. We cannot know in which order,
    # as 'genomes' is a dict, hence unordered, and as computation is done in parallel
    assert message_start_annot1 in messages
    assert message_start_annot2 in messages
    # messages Prokka cmd
    message_cmd1 = ("Prokka command: prokka --outdir test/data/annotate/generated_by_unit-tests/"
                    "H299_H561.fasta-prokkaRes --cpus 3 --prefix test_runall_1by1_1 "
                    "--centre prokka test/data/annotate/genomes/H299_H561.fasta")
    message_cmd2 = ("Prokka command: prokka --outdir test/data/annotate/generated_by_unit-tests/"
                    "genome1.fasta-prokkaRes --cpus 3 --prefix test_runall_1by1_2 "
                    "--centre prokka test/data/annotate/genomes/genome1.fasta")
    assert message_cmd1 in messages
    assert message_cmd2 in messages
    # Messages end annotation cmd
    message_end1 = ("End annotating test_runall_1by1_1 from "
                    "test/data/annotate/genomes/H299_H561.fasta.")
    message_end2 = ("End annotating test_runall_1by1_2 from "
                    "test/data/annotate/genomes/genome1.fasta.")
    assert message_end1 in messages
    assert message_end2 in messages
    # Messages error annotation cmd
    message_err1 = "test_runall_1by1_2 genome1.fasta: several .faa files"
    assert message_err1 in messages

